import os
import numpy as np
import scipy as sp
from numpy import ma
from collections import defaultdict
import itertools
import networkx as nx

from pymake import ExpTensor, GramExp, ExpeFormat, ExpSpace
from pymake.frontend.manager import ModelManager, FrontendManager

from pymake.util.algo import gofit, Louvain, Annealing
from pymake.util.math import reorder_mat, sorted_perm, categorical, clusters_hist
from pymake.util.utils import Now, nowDiff, colored
import matplotlib.pyplot as plt
from pymake.plot import plot_degree, degree_hist, adj_to_degree, plot_degree_poly, adjshow, plot_degree_2, random_degree,  draw_graph_circular, draw_graph_spectral, draw_graph_spring, _markers
from pymake.core.format import tabulate

from sklearn.metrics import roc_curve, auc, precision_recall_curve


class BurstProcess(ExpeFormat):

    _default_expe = dict(
        block_plot = False,
        _write  = False,
        _do            = ['burstiness', 'global'], # default
        _mode         = 'generative',
        gen_size      = 1000,
        epoch         = 30 , # Gen,eration epoch
        limit_gen   = 5, # Local superposition ! Ignored ?
        limit_class   = 15, # Ignored ?
    )

    def _preprocess(self):
        expe = self.expe

        frontend = FrontendManager.load(expe)
        if frontend:
            self._N = frontend.getN()
            expe.symmetric = frontend.is_symmetric()
        else:
            self._N = expe.N
            expe.symmetric = True


        if expe._mode == 'predictive':
            ### Generate data from a fitted model
            model = ModelManager.from_expe(expe, load=True)

            try:
                # this try due to mthod modification entry in init not in picke object..
                expe.hyperparams = model.get_hyper()
            except Exception as e:
                self.log.warning('loading hyperparam error: %s' % e)
                if model is not None:
                    model._mean_w = 0
                    expe.hyperparams = 0

        elif expe._mode == 'generative':
            ### Generate data from a un-fitted model

            expe.alpha = 1
            expe.gmma = 1/2
            expe.delta = [0.5,0.5]

            if 'ilfm' in expe.model:
                keys_hyper = ('alpha','delta')
                hyper = (expe.alpha, expe.delta)
            else:
                keys_hyper = ('alpha','gmma','delta')
                hyper = (expe.alpha, expe.gmma, expe.delta)
            expe.hyperparams = dict(zip(keys_hyper, hyper))
            expe.hyper = 'fix' # dummy
            model = ModelManager.from_expe(expe, load=False)

        else:
            raise NotImplementedError('What generation context ? predictive/generative..')

        self.log.info('=== GenNetworks === ')
        self.log.info('Mode: %s' % expe._mode)
        self.log.info('===')
        self.log.info('hyper: %s' % (str(expe.hyperparams)))

        self.frontend = frontend
        self.model = model

        if model is None:
            raise FileNotFoundError('No model for Expe at :  %s' % self.output_path)



    def _generate(self):
        ''' Generate data. '''
        expe = self.expe
        model = self.model

        Y = []
        Theta = []
        Phi = []
        now = Now()
        for i in range(expe.epoch):
            y, theta, phi = model.generate(mode=expe._mode,
                                           N=self._N, K=expe.K,
                                           symmetric=expe.symmetric,
                                          )
            Y.append(y)
            Theta.append(theta)
            Phi.append(phi)
        self.log.info('Data Generation : %s second' % nowDiff(now))

        #if expe.symmetric:
        #    for y in Y:
        #        frontend.symmetrize(y)

        self._Y = Y
        self._Theta = Theta
        self._Phi = Phi

    def _local_likelihood(self, theta, phi, k1, k2=None):
        ''' Expected local adjacency matrix more that likelihood. '''

        if not k2:
            k2 = k1

        model = self.expe.model

        if 'mmsb' in model:
            ll = np.outer(theta[:,k1], theta[:,k2]) * phi[k1, k2]
        elif 'ilfm' in model:
            ll = theta.dot(phi).dot(theta.T)
            ll = 1 / (1 + np.exp(-1 * ll))
            ll = sp.stats.bernoulli.rvs(ll) * np.outer(theta[:,k1], theta[:,k2])
        else:
            raise ValueError('Model unknow for local likelihood computation')
        return ll

    @ExpeFormat.raw_plot('corpus')
    def burst_process_me(self, frame):

        expe = self.expe

        # Force ONE epoch # bernoulli variance...
        expe.epoch = 1
        self._generate()

        Y = self._Y
        Theta = self._Theta
        Phi = self._Phi

        for i in range(expe.epoch):

            theta = Theta[i]
            phi = Phi[i]
            N = theta.shape[0]
            process = np.zeros((N, N))

            likelihood = self.model.likelihood(theta, phi)
            adj = sp.stats.bernoulli.rvs(likelihood)
            process = np.cumsum(adj, 1)


        ax = frame.ax()
        legend = expe.model
        step = 3
        ax.errorbar(np.arange(N)[0::step],
                    process.mean(0)[0::step],
                    yerr=process.std(0)[0::step],
                    fmt='o', label=legend,
                    elinewidth=1, capsize=0, alpha=0.3,
                   )

        ax.legend(loc=2,prop={'size':10})
        ax.set_xlabel('Epoch')
        ax.set_ylabel('Cumulative Sum')

    @ExpeFormat.raw_plot('corpus')
    def burst_process_mg(self, frame):

        expe = self.expe

        self._generate()

        Y = self._Y
        Theta = self._Theta
        Phi = self._Phi

        processes = []

        for i in range(expe.epoch):

            theta = Theta[i]
            phi = Phi[i]
            N = theta.shape[0]

            process = np.zeros((N, N))
            likelihood = self.model.likelihood(theta, phi)
            adj = sp.stats.bernoulli.rvs(likelihood)

            process = np.cumsum(adj, 1)
            processes.append(process)

        # Mean-Variance of the count process expectation
        mean_process = np.vstack((p.mean(0) for p in processes))
        # Std-Variance of the count process expectation
        var_process = np.vstack((p.var(0) for p in processes))

        legend = expe.model
        step = 3

        ax1 = frame.fig.add_subplot(1,2,1)
        ax2 = frame.fig.add_subplot(1,2,2)

        ax1.errorbar(np.arange(N)[0::step],
                    mean_process.mean(0)[0::step],
                    yerr=mean_process.std(0)[0::step],
                    fmt='o', label=legend,
                    elinewidth=1, capsize=0, alpha=0.3,)

        ax2.errorbar(np.arange(N)[0::step],
                    var_process.mean(0)[0::step],
                    yerr=var_process.std(0)[0::step],
                    fmt='o', label=legend,
                    elinewidth=1, capsize=0, alpha=0.3,)

        ax1.legend(loc=2,prop={'size':10})
        ax1.set_xlabel('Epoch')
        ax1.set_ylabel('Expectation')
        ax2.set_ylabel('variance')
        frame.title = 'degree Cumutative Sum'

        default_size = frame.fig.get_size_inches()
        frame.fig.set_size_inches((default_size[0]*2, default_size[1]))


    @ExpeFormat.raw_plot('model')
    def burst_process_local_me(self, frame):

        expe = self.expe

        # Force ONE epoch # bernoulli variance...
        expe.epoch = 1
        self._generate()

        Y = self._Y
        Theta = self._Theta
        Phi = self._Phi

        for i in range(expe.epoch):

            theta = Theta[i]
            phi = Phi[i]

            N = theta.shape[0]
            K = theta.shape[1]
            process = np.zeros((K, N, N))

            for _k in range(K):
                process[_k] = self._local_likelihood(theta, phi, _k)

            process = np.cumsum(process, 2)

        ax = frame.ax()
        step = 3


        for _k in range(K):
            _process = process[_k]
            legend = expe.model + ' K=%d'%(_k)
            ax.errorbar(np.arange(N)[0::step],
                        _process.mean(0)[0::step],
                        yerr=_process.std(0)[0::step],
                        fmt='o', label=legend,
                        elinewidth=1, capsize=0, alpha=0.3,
                       )

        ax.legend(loc=2,prop={'size':10})
        ax.set_xlabel('Epoch')
        ax.set_ylabel('Cumulative Sum')


    @ExpeFormat.raw_plot('model')
    def burst_process_local_mg(self, frame):

        expe = self.expe

        self._generate()

        Y = self._Y
        Theta = self._Theta
        Phi = self._Phi

        Ks = [o.shape[1] for o in Theta]
        k_min = min(Ks)
        k_max= max(Ks)

        processes = []

        for i in range(expe.epoch):

            theta = Theta[i]
            phi = Phi[i]

            N = theta.shape[0]
            K = theta.shape[1]
            #if K > k_min:
            #    print('warning: not all block converted.')
            #    continue
            process = np.zeros((K, N, N))

            for _k in range(K):
                process[_k] = self._local_likelihood(theta, phi, _k)

            if K < k_max:
                # Ks are ordered !
                _pad = k_max - K
                process = np.pad(process, [(0,_pad),(0,0),(0,0)], mode='constant', constant_values=0)

            process = np.cumsum(process, 2)
            processes.append(process)

        # Mean-Variance of the count process expectation
        mean_process = np.stack((p.mean(1) for p in processes))
        # Std-Variance of the count process expectation
        var_process = np.stack((p.var(1) for p in processes))

        ax1 = frame.fig.add_subplot(1,2,1)
        ax2 = frame.fig.add_subplot(1,2,2)
        step = 3

        for _k in range(K):
            _process = mean_process[:,_k]
            legend = expe.model + ' K=%d'%(_k)
            ax1.errorbar(np.arange(N)[0::step],
                        _process.mean(0)[0::step],
                        yerr=_process.std(0)[0::step],
                        fmt='o', label=legend,
                        elinewidth=1, capsize=0, alpha=0.3,)

            _process = var_process[:,_k]
            legend = expe.model + ' K=%d'%(_k)
            ax2.errorbar(np.arange(N)[0::step],
                        _process.mean(0)[0::step],
                        yerr=_process.std(0)[0::step],
                        fmt='o',# label=legend,
                        elinewidth=1, capsize=0, alpha=0.3,)

        #ax1.set_xscale('log'); ax1.set_yscale('log')
        #ax2.set_xscale('log'); ax2.set_yscale('log')
        ax1.legend(loc=2,prop={'size':10})
        ax1.set_xlabel('Epoch')
        ax1.set_ylabel('Expectation')
        ax2.set_ylabel('Variance')
        frame.title = 'Degree Cumulative Sum'

        default_size = frame.fig.get_size_inches()
        frame.fig.set_size_inches((default_size[0]*2, default_size[1]))


#
#
# Second phase ("temporal process")
#
#
#
    @ExpeFormat.raw_plot()
    #@ExpeFormat.raw_plot('corpus')
    def prop_process_me(self, frame, p=250):
        p = int(p)
        expe = self.expe

        # Force ONE epoch # bernoulli variance...
        expe.epoch = 1
        self._generate()

        Y = self._Y
        Theta = self._Theta
        Phi = self._Phi

        theta = Theta[0]
        phi = Phi[0]

        N = theta.shape[0]

        likelihood = self.model.likelihood(theta, phi)
        adj = sp.stats.bernoulli.rvs(likelihood)

        n_to_zeros = N-p
        _idx = np.random.choice(np.arange(N**2), n_to_zeros, replace=False)
        idx = np.unravel_index(_idx, (N,N))

        adj_n = adj.copy()
        adj_n[idx] = 0

        g_n = nx.from_numpy_array(adj_n)
        degree_n = dict(g_n.degree())
        d_n, dc_n = degree_hist(degree_n, filter_zeros=False)

        g = nx.from_numpy_array(adj)
        degree = dict(g.degree())
        d, dc = degree_hist(degree, filter_zeros=False)


        x = d_n
        y = []
        burstiness = defaultdict(lambda:0)
        normalize_deg = defaultdict(lambda:0)
        # Compute p(d(N)>n+1 | p(p)=n)
        for node, deg_n in degree_n.items():
            if deg_n == 0:
                continue

            deg_N = degree[node]
            if deg_N > deg_n:
                burstiness[deg_n] += 1

            normalize_deg[deg_n] +=1

        # Normalize
        for deg, total in normalize_deg.items():
            burstiness[deg] = burstiness[deg] / total

        for deg in x:
            y.append(burstiness[deg])

        ax = frame.ax()
        ax.plot(x, y, label=expe.model)

        ax.legend(loc=1,prop={'size':10})
        ax.set_xlabel('n')
        ax.set_ylabel('Cumulative Sum')
        frame.title = expe.corpus + ' p=%s' % p


    @ExpeFormat.raw_plot()
    #@ExpeFormat.raw_plot('corpus')
    def prop_process_local_me(self, frame, p=250):
        p = int(p)
        expe = self.expe

        # Force ONE epoch # bernoulli variance...
        expe.epoch = 1
        self._generate()

        Y = self._Y
        Theta = self._Theta
        Phi = self._Phi

        theta = Theta[0]
        phi = Phi[0]

        N = theta.shape[0]
        K = theta.shape[1]

        for _k1 in range(K):
            #for _k2 in range(K):

                n_to_zeros = N-p

                adj = self._local_likelihood(theta, phi, _k1)
                #if adj.dtype == np.dtype(float):
                #    adj = sp.stats.bernoulli.rvs(adj)

                #    _id1 = np.arange(N**2).reshape((N,N))[adj==1]
                #    _id0 = np.arange(N**2).reshape((N,N))[adj==0]
                #    nn1 = len(_id1) // 3
                #    nn0 = n_to_zeros - nn1
                #    if nn1 > 0:
                #        _idx1 = np.random.choice(_id1, nn1, replace=False)
                #        _idx0 = np.random.choice(_id0, nn0, replace=False)
                #        _idx = np.hstack((_idx0, _idx1))
                #    else:
                #        _idx = _id0

                #    idx = np.unravel_index(_idx, (N,N))
                _idx = np.random.choice(np.arange(N**2), n_to_zeros, replace=False)
                idx = np.unravel_index(_idx, (N,N))


                adj_n = adj.copy()
                adj_n[idx] = 0

                if 'ilfm' in expe.model:
                    g_n = nx.from_numpy_array(adj_n)
                    degree_n = dict(g_n.degree())
                else:
                    degree_n = dict((i,int(round(d))) for i,d in enumerate(adj_n.sum(1)))
                d_n, dc_n = degree_hist(degree_n, filter_zeros=False)

                if 'ilfm' in expe.model:
                    g = nx.from_numpy_array(adj)
                    degree = dict(g.degree())
                else:
                    degree = dict((i,int(round(d))) for i,d in enumerate(adj.sum(1)))
                d, dc = degree_hist(degree, filter_zeros=False)

                x = d_n
                y = []
                burstiness = defaultdict(lambda:0)
                normalize_deg = defaultdict(lambda:0)
                # Compute p(d(N)>n+1 | p(p)=n)
                for node, deg_n in degree_n.items():
                    if deg_n == 0:
                        continue

                    deg_N = degree[node]
                    if deg_N > deg_n:
                        burstiness[deg_n] += 1

                    normalize_deg[deg_n] +=1

                # Normalize
                for deg, total in normalize_deg.items():
                    burstiness[deg] = burstiness[deg] / total

                for deg in x:
                    y.append(burstiness[deg])

                ax = frame.ax()

                label = '%s K=%d' % (expe.model, _k1)
                ax.plot(x, y, label=label)

        ax.legend(loc=1,prop={'size':8})
        ax.set_xlabel('n')
        ax.set_ylabel('Cumulative Sum')
        frame.title = expe.corpus + ' p=%s' % p



    @ExpeFormat.raw_plot()
    #@ExpeFormat.raw_plot('corpus')
    def prop2_process_me(self, frame, p=90):
        p = int(p)
        expe = self.expe

        # Force ONE epoch # bernoulli variance...
        expe.epoch = 1
        self._generate()

        Y = self._Y
        Theta = self._Theta
        Phi = self._Phi

        theta = Theta[0]
        phi = Phi[0]

        N = theta.shape[0]

        likelihood = self.model.likelihood(theta, phi)
        adj = sp.stats.bernoulli.rvs(likelihood)

        n_to_zeros = int( N**2 * (1-p/100))
        _idx = np.random.choice(np.arange(N**2), n_to_zeros, replace=False)
        idx = np.unravel_index(_idx, (N,N))

        adj_n = adj.copy()
        adj_n[idx] = 0

        g_n = nx.from_numpy_array(adj_n)
        degree_n = dict(g_n.degree())
        d_n, dc_n = degree_hist(degree_n, filter_zeros=False)

        g = nx.from_numpy_array(adj)
        degree = dict(g.degree())
        d, dc = degree_hist(degree, filter_zeros=False)


        x = d_n
        y = []
        burstiness = defaultdict(lambda:0)
        normalize_deg = defaultdict(lambda:0)
        # Compute p(d(N)>n+1 | p(p)=n)
        for node, deg_n in degree_n.items():
            if deg_n == 0:
                continue

            deg_N = degree[node]
            if deg_N > deg_n:
                burstiness[deg_n] += 1

            normalize_deg[deg_n] +=1

        # Normalize
        for deg, total in normalize_deg.items():
            burstiness[deg] = burstiness[deg] / total

        for deg in x:
            y.append(burstiness[deg])

        ax = frame.ax()

        w = 0.2
        opacity = 0.8
        y = np.array(y)
        index = 0
        ticks = []
        ticks_label = []
        c = self.colors.next()
        for n, v in enumerate(y):
            if v == 0:
                continue
            rects = ax.bar(index + w, v, w,
                           alpha=opacity,
                           label=None, color=c)
            ticks.append(index)
            ticks_label.append(x[n])
            index += 1

        ax.set_xticklabels(ticks_label)
        ticks = np.array(ticks) + w
        ax.set_xticks(ticks)
        ax.xaxis.set_tick_params(labelsize=8)

        ax.legend(loc=4,prop={'size':10})
        ax.set_xlabel('n')
        ax.set_ylabel('Probability of new links')
        frame.title = '%s, %s,  p=%s' % (self.specname(expe.corpus),
                                         self.specname(expe.model), p)

    @ExpeFormat.raw_plot()
    def prop2_process_local_me(self, frame, p=90):
        p = int(p)
        expe = self.expe

        # Force ONE epoch # bernoulli variance...
        expe.epoch = 1
        self._generate()

        Y = self._Y
        Theta = self._Theta
        Phi = self._Phi

        theta = Theta[0]
        phi = Phi[0]

        N = theta.shape[0]
        K = theta.shape[1]

        ticks = []
        ticks_label = []

        nb_class = 0
        for _k1 in range(K):
            if 'ilfm' in expe.model:
                K =1
            for _k2 in range(K):
                if 'ilfm' in expe.model:
                    _k2 = _k1

                if nb_class >= 4:
                    break

                n_to_zeros = int( N**2 * (1-p/100))

                adj = self._local_likelihood(theta, phi, _k1, _k2)
                #if adj.dtype == np.dtype(float):
                #    adj = sp.stats.bernoulli.rvs(adj)

                #    _id1 = np.arange(N**2).reshape((N,N))[adj==1]
                #    _id0 = np.arange(N**2).reshape((N,N))[adj==0]
                #    nn1 = len(_id1) // 3
                #    nn0 = n_to_zeros - nn1
                #    if nn1 > 0:
                #        _idx1 = np.random.choice(_id1, nn1, replace=False)
                #        _idx0 = np.random.choice(_id0, nn0, replace=False)
                #        _idx = np.hstack((_idx0, _idx1))
                #    else:
                #        _idx = _id0

                #    idx = np.unravel_index(_idx, (N,N))
                _idx = np.random.choice(np.arange(N**2), n_to_zeros, replace=False)

                idx = np.unravel_index(_idx, (N,N))

                adj_n = adj.copy()
                adj_n[idx] = 0

                if 'ilfm' in expe.model:
                    g_n = nx.from_numpy_array(adj_n)
                    degree_n = dict(g_n.degree())
                else:
                    degree_n = dict((i,int(round(d))) for i,d in enumerate(adj_n.sum(1)))
                d_n, dc_n = degree_hist(degree_n, filter_zeros=False)

                if 'ilfm' in expe.model:
                    g = nx.from_numpy_array(adj)
                    degree = dict(g.degree())
                else:
                    degree = dict((i,int(round(d))) for i,d in enumerate(adj.sum(1)))
                d, dc = degree_hist(degree, filter_zeros=False)

                x = d_n

                y = []
                burstiness = defaultdict(lambda:0)
                normalize_deg = defaultdict(lambda:0)
                # Compute p(d(N)>n+1 | p(p)=n)
                for node, deg_n in degree_n.items():
                    if deg_n == 0:
                        continue

                    deg_N = degree[node]
                    if deg_N > deg_n:
                        burstiness[deg_n] += 1

                    normalize_deg[deg_n] +=1

                # Normalize
                for deg, total in normalize_deg.items():
                    burstiness[deg] = burstiness[deg] / total

                for deg in x:
                    y.append(burstiness[deg])

                y = np.array(y)
                if len(y[y>0]) <= 3:
                    continue

                ax = frame.ax()

                w = 0.4
                opacity = 0.8
                y = np.array(y)
                _index = 0
                index = _index + len(ticks)*1.25
                c = self.colors.next()
                label = '%s K=%d' % (expe.model, _k1)
                one_count = 0
                first = True
                for n, v in enumerate(y):
                    if v == 0:
                        continue
                    if v == 1:
                        one_count += 1
                    else:
                        one_count = 0

                    if one_count >=4:
                        break

                    if first:
                        label = 'classe %s' % nb_class
                    else:
                        label = None


                    rects = ax.bar(index + w/2, v, w,
                                   alpha=opacity,
                                   label=label, color=c)
                    ticks.append(index)
                    ticks_label.append(x[n])
                    index += 1
                    first = False

                nb_class +=1

        ax.set_xticklabels(ticks_label)
        ticks = np.array(ticks) + w
        ax.set_xticks(ticks)
        ax.xaxis.set_tick_params(labelsize=8)

        ax.legend(loc=4,prop={'size':10})
        ax.set_xlabel('n')
        ax.set_ylabel('Probability of new links')
        frame.title = '%s, %s,  p=%s' % (self.specname(expe.corpus),
                                         self.specname(expe.model), p)

    @ExpeFormat.raw_plot()
    def power_law_mm(self, frame, p=90):
        p = int(p)
        expe = self.expe

        # Force ONE epoch # bernoulli variance...
        expe.epoch = 1
        self._generate()

        Y = self._Y
        Theta = self._Theta
        Phi = self._Phi

        theta = Theta[0]
        phi = Phi[0]

        N = theta.shape[0]
        K = theta.shape[1]

        ticks = []
        ticks_label = []

        alpha = np.exp(self.model.s.zsampler.log_alpha_beta)[:K]

        for _i in range(N):

            if _i >= 3:
                break
            plt.figure()

            fi_params = (2*N + alpha.sum()) * theta[_i]

            samples = np.random.dirichlet(fi_params, 1000)

            for _k in range(K):
                plt.hist(samples[:, _k], bins=100)


            plt.ylim(0, 100)
            plt.xlabel('f_i %d values'%_i)
            plt.ylabel('count')







import sys, os
from collections import defaultdict
from string import Template

from .frontend import DataBase
from pymake.util.vocabulary import Vocabulary



class frontendText(DataBase):
    """ Frontend for text Corpus """

    def __init__(self, expe=None):
        super(frontendText, self).__init__(expe)
        self._data_type = 'text'

    def load_data(self, randomize=False):
        """ Load data according to different scheme:
            * Corpus from file dataset
            * Corpus from random generator
            """

        corpus_name = self.corpus_name
        self.get_corpus(corpus_name)

        # @DEBUG
        if str(self.N).isdigit() and int(self.N) > self.data.shape[0]:
            raise ValueError('Sampling of corpus %s too big (-n options)' % self.N)

        if randomize:
            self.shuffle_docs()
        return self.data

    def make_testset(self, ratio):
        self.log.warning('check WHY and WHEN overflow in stirling matrix !?')
        self.log.warning('debug why error and i get walue superior to 6000 in the striling matrix ????')
        D = self.data.shape[0]
        d = int(D * ratio)
        data = self.data[:d]
        data_t = self.data[d:]
        return data, data_t

    # @Debug: remove docu for wich count > 6000 because of stirling matrix !
    def sample(self, N=None, **args):
        N = N or self.N
        n = self.data.shape[0]
        if not N or N == 'all':
            self.N = 'all'
            # To remove !
            if  self.corpus_name == '20ngroups':
                data = self.data[:10000]
                empty_words =  np.where(data.sum(0).A[0] == 0)[0]
                new_cols = np.delete(np.arange(data.shape[1]), empty_words)
                self.data = data[:, new_cols]

        else:
            N = int(N)
            data = self.data[:N]
            # Here we come to streaming problem !
            # @DEBUG manage id2word
            if type(data) is np.array:
                empty_words =  np.where(data.sum(0) == 0)[0]
                self.data = np.delete(data, empty_words, axis=1)
            elif data.format == 'csr':
                empty_words =  np.where(data.sum(0).A[0] == 0)[0]
                new_cols = np.delete(np.arange(data.shape[1]), empty_words)
                self.data = data[:, new_cols]

        # @debug to remove
        _l =  (self.data >= 6000).sum(1).A.T[0]
        print(_l)
        tt = self.data[_l > 0]
        for t in tt:
            print(t[t>=6000])
        self.data = self.data[ _l == 0 ]
        return self.data

    ### Get and prepropress text
    #   See Vocabulary class...
    #   * Tokenisation from scratch
    #   * Stop Word from scratch
    #   * Lemmatization from Wornet
    #   * Load or Save in a Gensim context
    #       - Load has priority over Save
    # @Debug: There is a convertion to gensim corpus to use the serialization library and then back to scipy corpus.
    #   Can be avoided by using our own library of serialization, using Gensim if needed ?!
    def textloader(self, target, bdir=None, corpus_name="", n=None):
        if type(target) is str and os.path.isfile(target):
            bdir = os.path.dirname(target)
        elif bdir is None:
            bdir = self.basedir
        fn = 'corpus'
        if n:
            fn += str(n)
        elif type(target) is not str:
            n = len(target)
            fn += str(n)

        if corpus_name:
            fname = bdir + '/'+fn+'_' + corpus_name + '.mm'
        else:
            fname = bdir + '/'+fn+'.mm'

        if self._load_data and os.path.isfile(fname):
            data = gensim.corpora.MmCorpus(fname)
            data = gensim.matutils.corpus2csc(data, dtype=int).T
            id2word = dict(gensim.corpora.dictionary.Dictionary.load_from_text(fname + '.dico'))
        else:
            prin('re-Building Corpus...')
            raw_data, id2word = Vocabulary.parse_corpus(target)

            # Corpus will be in bag of words format !
            if type(raw_data) is list:
                voca = Vocabulary(exclude_stopwords=True)
                data = [voca.doc2bow(doc) for doc in raw_data]
                data = gensim.matutils.corpus2csc(data, dtype=int).T # Would be faster with #doc #term #nnz
            else:
                data = raw_data

            if self._save_data:
                make_path(bdir)
                _data = gensim.matutils.Sparse2Corpus(data, documents_columns=False)
                voca_gensim = gensim.corpora.dictionary.Dictionary.from_corpus(_data, id2word)
                voca_gensim.save_as_text(fname+'.dico')
                gensim.corpora.MmCorpus.serialize(fname=fname, corpus=_data)
                #@Debug how to get the corpus from list of list ?
                #_data = gensim.corpora.MmCorpus(fname)

        return data, id2word

    def get_corpus(self, corpus_name):
        self.make_io_path()
        bdir = self.basedir
        data_t = None
        data_t = None
        if corpus_name == 'random':
            # mmmh speak !
            data = self.random()
        if corpus_name == 'lucene':
            raise NotImplementedError
            #searcher = warm_se(config)
            #q = config.get('q'); q['limit'] = config['limit_train']
            #id2word = searcher.get_id2word()
            #corpus = searcher.self.parse_corpus(q, vsm=config['vsm'], chunk=1000, batch=True)
        elif corpus_name == '20ngroups_sklearn':
            from sklearn.datasets import fetch_20newsgroups
            ngroup_train = fetch_20newsgroups(subset='train', remove=('headers', 'footers', 'quotes'), categories=None)
            ngroup_test = fetch_20newsgroups(subset='test', remove=('headers', 'footers', 'quotes'), categories=None)
            train_data = ngroup_train.data
            test_data = ngroup_test.data
            #corpus, id2word = self.textloader(train_data, bdir=bdir, corpus_name='train', n=config.get('N'))
            corpus, id2word = self.textloader(train_data, bdir=bdir, corpus_name='train')
            corpus_t, id2word_t = self.textloader(test_data, bdir=bdir, corpus_name='test')

            K = self.K
            #################
            ### Group Control
            test_classes = ngroup_test.target
            train_classes = ngroup_train.target
            if K == 6 and len(ngroup_test.target_names) != 6:
                # Wrap to subgroups
                target_names = ['comp', 'misc', 'rec', 'sci', 'talk', 'soc']
                map_ = dict([(0,5), (1,0), (2,0), (3,0), (4,0), (5,0), (6,1), (7,2), (8,2), (9,2), (10,2), (11,3), (12,3), (13,3), (14,3), (15,5), (16,4), (17,4), (18,4), (19,5)])
                test_classes = set_v_to(test_classes, map_)
                train_classes = set_v_to(train_classes, map_)
            else:
                target_names = ngroup_test.target_names
            C = len(target_names)

        elif corpus_name == 'wikipedia':
            # ? file type
            # Create
            command = './gensim/gensim/scripts/make_wikicorpus_ml.py '
            command += '/work/data/wikipedia/enwiki-latest-pages-articles.xml.bz2 ../pymake/data/wikipedia/wiki_en'
            os.system(command)
            # Load
            error = 'Load Wikipedia corpus'
            raise NotImplementedError(error)
        elif corpus_name == 'odp':
            # SVMlight file type
            from sklearn.datasets import load_svmlight_files, load_svmlight_file
            fn_train = os.path.join(bdir, 'train.txt')
            fn_test = os.path.join(bdir, 'test.txt')
            # More feature in test than train !!!
            data, train_classes = load_svmlight_file(fn_train)
            data_t, test_classes = load_svmlight_file(fn_test)
            id2word = None
        elif corpus_name in ('reuter50', 'nips12', 'nips', 'enron', 'kos', 'nytimes', 'pubmed') or corpus_name == '20ngroups' :
            # DOC_ID FEAT_ID COUNT file type
            data, id2word = self.textloader(bdir, corpus_name=corpus_name)
        else:
            raise ValueError('Which corpus to Load ?')

        self.data = data
        self.id2word = id2word
        if data_t is None:
            pass
            #raise NotImplementedError('Corpus test ?')
        else:
            self.data_t = data_t

        return True

    def get_data_prop(self):
        prop = defaultdict()
        prop.update( {'corpus': self.corpus_name,
                      'instances' : self.data.shape[1] })
        nnz = self.data.sum()
        _nnz = self.data.sum(axis=1)
        dct = {'features': self.data.shape[1],
               'nnz': nnz,
               'nnz_mean': _nnz.mean(),
               'nnz_var': _nnz.var(),
               'train_size': None,
               'test_size': None,
              }
        prop.update(dct)
        return prop

    def template(self, dct):
        text_templ = '''###### $corpus_name
        Building: $time minutes
        Documents: $instances
        Nnz: $nnz
        Nnz mean: $nnz_mean
        Nnz var: $nnz_var
        Vocabulary: $features
        train: $train_size
        test: $test_size
        \n'''
        return Template(templ).substitute(dct)

    def print_vocab(self, data, id2word):
        if id2word:
            return gensim.corpora.dictionary.Dictionary.from_corpus(data, id2word) #; print voca

    def shuffle_docs(self):
        self.shuffle_instances()

    # Return a vector with document generated from a count matrix.
    # Assume sparse matrix
    @staticmethod
    def sparse2stream(data):
        #new_data = []
        #for d in data:
        #    new_data.append(d[d.nonzero()].A1)
        bow = []
        for doc in data:
            # Also, see collections.Counter.elements() ...
            bow.append( np.array(list(chain(* [ doc[0,i]*[i] for i in doc.nonzero()[1] ]))))
        bow = np.array(bow)
        #map(np.random.shuffle, bow)
        return bow


    # Debug
    def run_lda(self):
        pass
   #     # Cross Validation settings...
   #     #@DEBUG: do we need to remake the vocabulary ??? id2word would impact the topic word distribution ?
   #     if corpus_t is None:
   #         pass
   #         #take 80-20 %
   #         # remake vocab and shape !!!
   #         # manage downside
   #     try:
   #         total_corpus = len(corpus)
   #         total_corpus_t = len(corpus_t)
   #     except:
   #         total_corpus = corpus.shape[0]
   #         total_corpus_t = corpus.shape[0]
   #     if config.get('N'):
   #         N = config['N']
   #     else:
   #         N = total_corpus
   #     corpus = corpus[:N]
   #     n_percent = float(N) / total_corpus
   #     n_percent = int(n_percent * total_corpus_t) or 10
   #     heldout_corpus = corpus_t[:n_percent]

   #     ############
   #     ### Load LDA
   #     load = config['load_model']
   #     # Path for LDA model!
   #     bdir = '../PyNPB/data/'
   #     bdir = os.path.join(bdir,config.get('corpus'), config.get('bdir', ''))
   #     lda = lda_gensim(corpus, id2word=id2word, K=K, bdir=bdir, load=load, model=config['model'], alpha=config['hyper'], n=config['N'], heldout_corpus=heldout_corpus)
   #     lda.inference_time = datetime.now() - last_d
   #     last_d = ellapsed_time('LDA Inference -- '+config['model'], last_d)

   #     ##############
   #     ### Log Output
   #     lda.print_topics(K)

   #     ##############
   #     ### Prediction
   #     corpus_t = corpus
   #     if config['predict'] and true_classes is not None and C == K:
   #         true_classes = train_classes
   #         predict_class = []
   #         confusion_mat = np.zeros((K,C))
   #         startt = datetime.now()
   #         for i, d in enumerate(corpus_t):
   #             d_t = lda.get_document_topics(d, minimum_probability=0.01)
   #             t = max(d_t, key=lambda item:item[1])[0]
   #             predict_class.append(t)
   #             c = true_classes[i]
   #             confusion_mat[t, c] += 1
   #         last_d = ellapsed_time('LDA Prediction', startt)
   #         predict_class = np.array(predict_class)
   #         lda.confusion_matrix = confusion_mat

   #         map_kc = map_class2cluster_from_confusion(confusion_mat)
   #         #new_predict_class = set_v_to(predict_class, dict(map_kc))

   #         print "Confusion Matrix, KxC:"
   #         print confusion_mat
   #         print map_kc
   #         print [(k, target_names[c]) for k,c in map_kc]

   #         purity = confusion_mat.max(axis=1).sum() / len(corpus_t)
   #         print 'Purity (K=%s, C=%s, D=%s): %s' % (K, C, len(corpus_t), purity)

   #         #precision = np.sum(new_predict_class == true_classes) / float(len(predict_class)) # equal !!!
   #         precision = np.sum(confusion_mat[zip(*map_kc)]) / float(len(corpus_t))
   #         print 'Ratio Groups Control: %s' % (precision)

   #     if save:
   #         ## Too big
   #         lda.expElogbeta = None
   #         lda.sstats = None
   #         lda.save(lda.fname)

   #     if config.get('_verbose'):
   #         #print lda.top_topics(corpus)
   #         for d in corpus:
   #             print lda.get_document_topics(d, minimum_probability=0.01)

   #     print lda
   #     if type(corpus) is not list:
   #         print corpus
   #         print corpus_t
   #     self.print_vocab(corpus, id2word)





from copy import copy, deepcopy
import traceback,  importlib
import numpy as np
from collections import OrderedDict, defaultdict
from itertools import product

from pymake import get_pymake_settings
from pymake.util.utils import get_dest_opt_filled, hash_objects, ask_sure_exit, basestring
from pymake.index.indexmanager import IndexManager as IX

import logging
lgg = logging.getLogger('root')


''' Structure of Pymake Objects.
'''


from tabulate import tabulate

# Ugly, integrate.
def _table_(tables, headers=[], max_line=10, max_row=30, name=''):

    if isinstance(headers, str):
        # tables is dict
        # Sort the dict
        ordered_keys = sorted(tables.keys())
        tables = OrderedDict([(k,tables[k]) for k in ordered_keys ])

        _tables = []
        cpt = 0
        max_row = 10
        for k, v in tables.items():
            if cpt % max_row == 0:
                t = OrderedDict()
                _tables.append(t)
            t[k] = v
            cpt += 1

        sep = '# %s'%name +  '\n'+'='*20
        print(sep)
        tables = '\n\n'.join([str(tabulate(t, headers=headers)) for t in _tables])
        return tables
    else:
        # tables is list
        raw = []
        for sec, table in enumerate(tables):
            table = sorted(table, key=lambda x:x[0])
            size = len(table)
            if size == 0:
                continue
            col = int((size-0.1) // max_line)
            junk = max_line % size
            table += ['-']*junk
            table = [table[j:max_line*(i+1)] for i,j in enumerate(range(0, size, max_line))]
            table = np.char.array(table).astype('|S'+str(max_row))
            fmt = 'simple'
            raw.append(tabulate(table.T,
                                headers=[headers[sec]]+['']*(col),
                                tablefmt=fmt))
        sep = '\n'+'='*20+'\n'
        return sep[1:] + sep.join(raw)




# Not sure this one is necessary, or not here
class BaseObject(object):
    ''' Notes : Avoid method conflict by ALWAYS settings this class in last
                at class definitions.
    '''

    def __init__(self, *args, **kwargs):
    #def __init__(self, name='BaseObject'):
        # Le ruban est infini...
        #if name is None:
        #    print(traceback.extract_stack()[-2])
        #    fn,ln,func,text = traceback.extract_stack()[-2]
        #    name = text[:text.find('=')].strip()
        pass

    #def _name(self):
    #    return self.__name__
    def items(self):
        return [(str(i), j) for i,j in enumerate(self)]
    def table(self):
        return tabulate(self.items())

class ExpSpace(dict):
    """ A dictionnary with dot notation access.
        Used for the **expe** settings stream.
    """

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__
    #__builtins__.hasattr = hasattr

    def __init__(self, *args, **kwargs):
        super(ExpSpace, self).__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    self[k] = v

        if kwargs:
            for k, v in kwargs.items():
                self[k] = v

    def copy(self):
        return type(self)(self)
    def __copy__(self):
        return self.__class__(**self)
    def __deepcopy__(self, memo):
        return self.copy()

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            # Default pmk settings
            if key == '_write':
                return False

            lgg.debug('an ExpSpace request exceptions occured for key: %s ' % (key))
            raise AttributeError(key)

    # Scratch method because __hasattr__ catch an error in getattr.
    def hasattr(self, key):
        return key in self

    # For Piclking
    def __getstate__(self):
        return self
    def __setstate__(self, state):
        self.update(state)
        self.__dict__ = self

class ExpVector(list, BaseObject):
    ''' A List of elements of an ExpTensor. '''
    def __add__(self, other):
        return self.__class__(list.__add__(self, other))
    def __sub__(self, other):
        return self.__class__([item for item in self if item not in other])

class ExpGroup(list, BaseObject):
    ''' A List of elements of an ExpTensor. '''

    def __init__(self, args, **kwargs):
        if kwargs:
            args = deepcopy(args)

        if isinstance(args, dict):
            args = [args]

        # Don't work well, why ?
        #for i, o in enumerate(args):
        #    if isinstance(o, (dict, ExpGroup)):
        #        args[i] = deepcopy(o)

        list.__init__(self, args)
        BaseObject.__init__(args, **kwargs)

        # Recursively update value if kwargs found.
        if len(kwargs) > 0:
            self.update_all(self, kwargs)

    def update_all(self, l, d):
        for o in l:
            if isinstance(o, list):
                self.update_all(o, d)
            elif isinstance(o, dict):
                for k, v in d.items():
                    o[k] = v
        return

    def __add__(self, other):
        return self.__class__(list.__add__(self, other))
    def __sub__(self, other):
        return self.__class__([item for item in self if item not in other])

class Spec(BaseObject):

    @staticmethod
    def find(spec, field='expe_name'):
        ix = IX(default_index='spec')
        spec = ix.getfirst(spec, field=field)
        return spec

    @staticmethod
    def get(scriptname, *expe):
        ix = IX(default_index='spec')
        raise NotImplementedError

    @staticmethod
    def get_all():
        ix = IX(default_index='spec')
        _res = ix.query(field='expe_name', terms='module_name')
        return _res

    @staticmethod
    def load(expe_name, expe_module):
        # debug to load from module or expe_name !

        p =  expe_module.split('.')
        modula, modulb = '.'.join(p[:-1]), p[-1]
        try:
            expdesign = getattr(importlib.import_module(modula), modulb)
            exp = getattr(expdesign, expe_name)
        except AttributeError as e:
            lgg.error("Seems that a spec `%s' has been removed : %s" % (expe_name, e))
            lgg.error("Fatal Error: unable to load spec:  try `pmk update' or try again.")
            exit(2)

        return exp, expdesign


    @classmethod
    def table(cls):
        ix = IX(default_index='spec')
        t = OrderedDict()
        for elt in ix.query(index='spec', terms=True):
            name = elt['module_name'].split('.')[-1]
            obj, _ = cls.load(elt['expe_name'], elt['module_name'])
            if isinstance(obj, (ExpSpace, ExpTensor, ExpGroup)):
                expes = t.get(name, []) + [ elt['expe_name'] ]
                t[name] = sorted(expes)
        return _table_(t, headers='keys', name=cls.__name__)

    # no more complex.
    # @sortbytype
    @classmethod
    def table_topos(cls, _spec):

        Headers = OrderedDict((('Corpuses', Corpus),
                               ('Models', Model),
                               ('Vector', ExpVector),
                               ('Exp', (ExpSpace, ExpTensor, ExpGroup)),
                               ('Unknown', str)))

        tables = [ [] for i in range(len(Headers))]

        for expe_name, expe_module in _spec.items():
            expe, _ = cls.load(expe_name, expe_module)
            try:
                pos = [isinstance(expe, T) for T in Headers.values()].index(True)
            except ValueError:
                pos = len(Headers) - 1
            tables[pos].append(expe_name)


        return _table_(tables, headers=list(Headers.keys()))



class Script(BaseObject):

    @staticmethod
    def get_all(_type='flat'):
        ix = IX(default_index='script')

        if _type == 'flat':
            _res = ix.query(field='method')
        elif _type == 'hierarchical':
            _res = ix.query(field='scriptsurname', terms=True)
        return _res

    @staticmethod
    def get(scriptname, arguments, field='scriptsurname'):

        ix = IX(default_index='script')
        topmethod = ix.getfirst(scriptname, field=field)
        if not topmethod:
            # get the first method that have this name
            topmethod = ix.getfirst(scriptname, field='method')
            if not topmethod:
                return None
                #try:
                #    raise ValueError('error: Unknown script: %s' % (scriptname))
                #except:
                #    # Exception from pyclbr
                #    # index commit race condition I guess.
                #    print('error: Unknown script: %s' % (scriptname))
                #    exit(42)

            arguments = [scriptname] + arguments
            #script_name = topmethod['scriptsurname']

        module = importlib.import_module(topmethod['module'])
        script = getattr(module, topmethod['scriptname'])
        return script, arguments

    @classmethod
    def table(cls):
        ix = IX(default_index='script')
        t = OrderedDict()
        for elt in  ix.query(index='script', terms=True):
            name = elt['scriptname']
            methods = t.get(name, []) + [ elt['method'] ]
            t[name] = sorted(methods)
        return _table_(t, headers='keys', name=cls.__name__)


class Corpus(ExpVector):

    # Meta-grammar / Ontology :
    #   Corpus := {
    #               data_type : text | network | image | ...,
    #               dtype : "specific to a data_type"
    #               name : "object identifier"
    #               path : "disk repo"
    #               source : url | random
    #             }

    # IX integration needed..

    _corpus_data = [
        dict(name='clique'        , data_type='network', data_source='random', directed=False),
        dict(name='generator'     , data_type='network', data_source='random', directed=False, nodes=1000),
        dict(name='graph'         , data_type='network', data_source='random'),
        dict(name='alternate'     , data_type='network', data_source='random', directed=False),
        dict(name='BA'            , data_type='network', data_source='random'),
        dict(name='manufacturing' , data_type='network', data_source='web', directed=True, nodes=167, edges=5784, density=0.209, weighted=True),
        dict(name='fb_uc'         , data_type='network', data_source='web', directed=True, nodes=1899, edges=22195, density=0.006, weighted=True),
        dict(name='blogs'         , data_type='network', data_source='web', directed=True, nodes=1490, edges=19025, density=0.009, weighted=False),
        dict(name='emaileu'       , data_type='network', data_source='web', directed=True, nodes=1005, edges=25571, density=0.025, weighted=False),
        dict(name='propro'        , data_type='network', data_source='web', directed=False, nodes=2113, edges=1432, density=0.001, weighted=False),
        dict(name='euroroad'      , data_type='network', data_source='web', directed=True, nodes=1177, edges=1432, density=0.001, weighted=False),

        # gt
        dict(name='astro-ph',    data_type='network', data_source='gt', directed=False, nodes=16706, edges=121251, weighted=True),
        dict(name='cond-mat',    data_type='network', data_source='gt', directed=False, nodes=16726, edges=47594 , weighted=True),
        dict(name='hep-th',      data_type='network', data_source='gt', directed=False, nodes=8361,  edges=15751 , weighted=True),
        dict(name='netscience',  data_type='network', data_source='gt', directed=False, nodes=1589,  edges=2742  , weighted=True),
        dict(name='email-Enron', data_type='network', data_source='gt', directed=False, nodes=36692, edges=367662, weighted=False), # time weighted

        #dict(name='facebook'     ,  data_type='network', data_source='web', directed=True, nodes=None, edges=None, density=None, wheigted=None),

        #dict(name='reuter50'  , data_type='text', data_source='web'),
        #dict(name='nips12'    , data_type='text', data_source='web'),
        #dict(name='nips'      , data_type='text', data_source='web'),
        #dict(name='enron'     , data_type='text', data_source='web'),
        #dict(name='kos'       , data_type='text', data_source='web'),
        #dict(name='nytimes'   , data_type='text', data_source='web'),
        #dict(name='pubmed'    , data_type='text', data_source='web'),
        #dict(name='20ngroups' , data_type='text', data_source='web'),
        #dict(name='odp'       , data_type='text', data_source='web'),
        #dict(name='wikipedia' , data_type='text', data_source='web'),
        #dict(name='lucene', data_type='text', data_source='lucene'), # needs field spec
        #dict(name='mongo', data_type='text', data_source='mongo'), # needs field spec
    ]

    @classmethod
    def get(cls, corpus_name):
        if not corpus_name:
            return None

        corpus = False

        # index/mongo...
        for data in cls._corpus_data:
            if corpus_name.startswith(data['name']):
                corpus = data.copy()
                break

        return corpus

    @classmethod
    def get_all(cls):
        return cls._corpus_data

class Model(ExpVector):

    @staticmethod
    def get(model_name):
        ix = IX(default_index='model')

        _model =  None
        docir = ix.getfirst(model_name, field='surname')
        if docir:
            mn = importlib.import_module(docir['module'])
            #mn = importlib.import_module(docir['module'], package=local_package)
            _model = getattr(mn, docir['name'], None)
        return _model

    @staticmethod
    def get_all(_type='short'):
        ix = IX(default_index='model')
        if _type == 'short':
            res = ix.query(field='surname')
        elif _type == 'topos':
            _res = ix.query(field='surname', terms=True)
            res = []
            for elt in _res:
                # beurk
                if len(elt['category']) > 0:
                    # means that len(surname.split('.')) > 1
                    names = elt['surname'].split('.')
                    topos = '.'.join(elt['category'].split())
                    surname = '.'.join((names[0],  topos , names[1]))
                else:
                    surname = elt['surname']
                res.append(surname)
        return res

    @classmethod
    def table(cls, _type='short'):
        tables = cls.get_all(_type),
        return _table_(tables, headers=['Models'])


class ExpTensor(OrderedDict, BaseObject):
    ''' Represent a set of Experiences (**expe**). '''
    def __init__(self,  *args, **kwargs):
        OrderedDict.__init__(self, *args, **kwargs)
        BaseObject.__init__(self)

        self._size = 0

    @classmethod
    def from_expe(cls, conf=None, expe=None, parser=None):
        ''' Return the tensor who is an OrderedDict of iterable.
            Assume conf is an exp. Non list value will be listified.

            Parameters
            ----------
            expe : (ExpDesign, ExpSpace or dict)
                A design of experiment.
        '''
        _conf = conf.copy()
        if expe is None:
            expe = conf

        if not issubclass(type(expe), (cls, ExpSpace, dict, ExpVector)):
            raise ValueError('Expe not understood: %s' % type(expe))

        if issubclass(type(expe), Corpus):
            tensor = cls(corpus=expe)
        elif issubclass(type(expe), Model):
            tensor = cls(model=expe)
        elif issubclass(type(expe), ExpVector):
            tensor = cls((str(i),j) for i,j in enumerate(expe))
        elif isinstance(expe, ExpTensor):
            tensor = expe.copy()
        elif isinstance(expe, (dict, ExpSpace)):
            tensor = cls()
            tensor.update_from_dict(expe)
        else:
            raise NotImplementedError('input type of ExpVector unknow %s' % (expe))

        for k, v in tensor.items():
            if not issubclass(type(v), (list, set, tuple)):
                tensor[k] = [v]

        if _conf:
            tensor.update_from_dict(_conf, parser=parser)

        return tensor

    def update_from_dict(self, d, parser=None):
        ''' Update a tensor from a dict

            Parameters
            ----------
            d : dict
                the dict that uptate the tensor
            from_argv : bool
                if True, the is assumed to come from an CLI argparser. if the following conds are true :
                    * the settings in {d} are specified in the CLI (@check already filtererd in GramExp.parseargs)
                    * the settings in {d} is not in the CLI, and not in self.

            Notes
            -----
            SHould inherit _reserved keyword to prevent
        '''

        if parser is not None:
            dests_filled = get_dest_opt_filled(parser)

        for k, v in d.items():
            if k in ['_id_expe']:
                continue

            if parser is not None:
                if not k in dests_filled and k in self :
                    continue

            if issubclass(type(v), ExpVector):
                self[k] = v
            else:
                self[k] = [v]

    def get_size(self, virtual=False):
        if virtual:
            return  np.prod([len(x) for x in self.values()])
        else:
            return self._size


    def push_dict(self, d):
        ''' push one dict inside a exptensor.
            It extend _bind rule to filter the tensor.
        '''
        tensor_len = np.prod([len(x) for x in self.values()])
        if len(self) == 0:
            self.update_from_dict(d)
            return True

        _need_bind = False
        _up_dict = {}
        for k, v in d.items():
            if k in ['_id_expe']:
                continue

            vector = self.get(k, []).copy()
            if v not in vector:
                if len(vector) == 0:
                    _need_bind = True
                    #lgg.debug('setting to bind: (%s : %s)' % (k, v))
                    break
                vector.append(v)
            _up_dict[k] = vector

        if _need_bind:
            #raise NotImplementedError('Need to push bind value to build a tensor from non-overlaping settings.')
            return False
        else:
            self.update(_up_dict)
            return True


    def table(self, extra=[]):
        return tabulate(extra+sorted(self.items(), key=lambda x:x[0]),
                               headers=['Params','Values'])

# @debug : Rename this class to ?
class ExpTensorV2(BaseObject):
    ''' Represent a set of Experiences (**expe**) of type ExpTensor... '''
    def __init__(self, private_keywords=[]):
        BaseObject.__init__(self)
        self._private_keywords = private_keywords

        # --- Those are aligned ---
        self._tensors = [] # list of ExpTensor
        self._bind = []
        self._null = defaultdict(list)
        self._hash = []
        self._ds_ = [] # ExpDesign class per tensor
        #
        self._lod = [] # list of dict
        self._ds = [] # ExpDesign class per expe
        # --- meta ---
        self._conf = {}
        self._size = None

    @classmethod
    def from_conf(cls, conf, _max_expe=2e6, private_keywords=[], expdesign=None):
        gt = cls(private_keywords=private_keywords)
        _spec = conf.pop('_spec', None)
        if not _spec:
            if not expdesign:
                expdesign = ExpDesign
            conf['_name_expe'] = '_default_expe'
            conf['_expe_hash'] = hash_objects(dict((k,v) for k,v in conf.items() if k not in private_keywords))
            gt._tensors.append(ExpTensor.from_expe(conf))
            gt._ds_.append(expdesign)
            return gt

        exp = []
        size_expe = len(_spec)
        consume_expe = 0
        while consume_expe < size_expe:
            o = _spec[consume_expe]
            if isinstance(o, tuple):
                #_type => expdesign
                name, o, _type = o

            if isinstance(o, ExpGroup):
                size_expe += len(o) -1
                _spec = _spec[:consume_expe] + o + _spec[consume_expe+1:]
            elif isinstance(o, list):
                exp.append(o)
                gt._ds_.append(_type)
                consume_expe += 1
            else:
                o['_name_expe'] = name
                o['_expe_hash'] = hash_objects(dict((k,v) for k,v in o.items() if k not in private_keywords))
                if hasattr(_type, '_alias'):
                    o['_alias'] = getattr(_type, '_alias')

                exp.append(o)
                gt._ds_.append(_type)
                consume_expe += 1

            if size_expe > _max_expe:
                lgg.warning('Number of experiences exceeds the hard limit of %d (please review ExpTensor).' % _max_expe)

        gt._tensors.extend([ExpTensor.from_expe(conf, spec) for spec in exp])
        return gt

    def __iter__(self):
        for tensor in self._tensors:
            yield tensor

    def __len__(self):
        return self.get_size()

    def remove_all(self, key):
        if key in self._conf:
            self._conf.pop(key)

        for tensor in self._tensors:
            if key in tensor:
                tensor.pop(key)

        # @Debug self._lod is left untouched...
        # Really ?
        for d in self._lod:
            if key in d:
                d.pop(key)

    def update_all(self, **kwargs):
        self._conf.update(kwargs)

        for tensor in self._tensors:
            tensor.update_from_dict(kwargs)

        for d in self._lod:
            d.update(kwargs)

    def set_default_all(self, defconf):
        ''' set default value in exp '''

        # Update current spec with _default_expe
        for k, v in defconf.items():

            for tensor in self._tensors:
                if not k in tensor:
                    tensor[k] = [v]
            for expe in self._lod:
                if not k in expe:
                    expe[k] = v
            if k in self._conf:
                # @debug: dont test if all the group have this unique value.
                self._conf[k] = v

    def get_all(self, key, default=[]):
        ''' Get all values associated to a given key. '''

        vec = []

        if hasattr(self, '_lod'):
            for d in self._lod:
                if key in d:
                    vec.append(d[key])
        else:
            for tensor in self._tensors:
                vec.extend(tensor.get(key, []))

        if not vec:
            return default
        else:
            return vec

    def get_nounique_keys(self):
        ''' Return key that has gridded (different value occurence in the set of tensor). '''
        keys = defaultdict(set)
        for tensor in self._tensors:
            for k in tensor:
                o = tensor.get(k, [])
                for v in o:
                    if isinstance(v, str) and not v.startswith('_'):
                        keys[k].add(v)

        nounique_keys = []
        for k, _set in keys.items():
            if len(_set) > 1:
                nounique_keys.append(k)

        return nounique_keys


    def get_conf(self):
        _conf = {}
        for tensor in self._tensors:
            for k, v in tensor.items():
                if len(v) != 1:
                    if k in _conf:
                        _conf.pop(k)
                    continue

                if k in _conf and v[0] != _conf[k]:
                    _conf.pop(k)
                    continue
                else:
                    _conf[k] = v[0]

            #_confs.append(_conf)

        self._conf = _conf
        return self._conf

    def get_size(self):
        size = 0
        for tensor in self._tensors:
            size += tensor.get_size()
        self._size = size
        return self._size

    def check_bind(self):
        ''' Rules Filter '''

        for tensor in self._tensors:

            if '_bind' in tensor:
                _bind = tensor.pop('_bind')
                if not isinstance(_bind, list):
                    _bind = [_bind]
                elif len(_bind) ==1 and isinstance(_bind[0], list):
                    _bind = _bind[0]
            else:
                #_bind = getattr(self, '_bind', [])
                _bind = []

            self._bind.append(_bind)

    def check_model_typo(self):
        ''' Assume default module is pymake '''
        for tensor in self._tensors:
            models = tensor.get('model', [])
            for i, m in enumerate(models):

                if not '.' in m:
                    # Set the model ref name
                    pkg = get_pymake_settings('default_model')
                    if len(pkg) > 8:
                        prefix = pkg[:3]
                        if '.' in pkg:
                            prefix  += ''.join(map(lambda x:x[0], pkg.split('.')[1:]))
                    else:
                        prefix = pkg.split('.')[0]

                    models[i] = '%s.%s'%(prefix, m)

    def check_null(self):
        ''' Filter _null '''
        for tensor in self._tensors:
            for k in list(tensor.keys()):
                if '_null' in tensor.get(k, []):
                    v = tensor.pop(k)
                    self._null[k].append(v)

    def make_lod(self, skip_check=False):
        ''' Make a list of Expe from tensor, with filtering '''

        self._lod = []
        for _id, tensor in enumerate(self._tensors):
            lods = self._make_lod(tensor, _id)
            tensor._size = len(lods)
            self._lod.extend(lods)
            self._ds.extend([self._ds_[_id]]*len(lods))

        self._make_hash(skip_check)
        return self._lod

    def _make_lod(self, tensor, _id):
        ''' 1. make dol to lod
            2. filter _bind rule
            3. add special parameter (expe_id)
        '''
        if len(tensor) == 0:
            lod =  []
        else:
            len_l = [len(l) for l in tensor.values()]
            keys = sorted(tensor)
            lod = [dict(zip(keys, prod)) for prod in product(*(tensor[key] for key in keys))]

        # POSTFILTERING
        # Bind Rules
        idtoremove = []
        for expe_id, d in enumerate(lod):
            for rule in self._bind[_id]:
                _bind = rule.split('.')
                values = list(d.values())

                # This is only for  last dot separator process
                for j, e in enumerate(values):
                    if type(e) is str:
                        values[j] = e.split('.')[-1]


                if len(_bind) == 2:
                    # remove all occurence if this bind don't occur
                    # simltaneous in each expe.
                    a, b = _bind
                    if b.startswith('!'):
                        # Exclusif Rule
                        b = b[1:]
                        if a in values and b in values:
                            idtoremove.append(expe_id)
                    else:
                        # Inclusif Rule
                        if a in values and not b in values:
                            idtoremove.append(expe_id)

                elif len(_bind) == 3:
                    # remove occurence of this specific key:value if
                    # it does not comply with this bind.
                    a, b, c = _bind
                    # Get the type of this key:value.
                    _type = type(d[b])
                    if _type is bool:
                        _type = lambda x: True if x in ['True', 'true', '1'] else False

                    if c.startswith('!'):
                        # Exclusif Rule
                        c = c[1:]
                        if a in values and _type(c) == d[b]:
                            idtoremove.append(expe_id)
                    else:
                        # Inclusif Rule
                        if a in values and _type(c) != d[b]:
                            idtoremove.append(expe_id)


        lod = [d for i,d in enumerate(lod) if i not in idtoremove]
        # Save true size of tensor (_bind remove)
        self._tensors[_id]._size = len(lod)

        # Add extra information in lod expes
        n_last_expe = sum([t._size for t in self._tensors[:_id]])
        for _id, expe in enumerate(lod):
            expe['_id_expe'] = _id + n_last_expe

        return lod

    # @todo; lhs for clustering expe applications.
    def _make_hash(self, skip_check=False):
        _hash = []
        n_duplicate = 0
        for _id, _d in enumerate(self._lod):
            d = _d.copy()
            [ d.pop(k) for k in self._private_keywords if k in d and k != '_repeat']
            o = hash_objects(d)
            if o in _hash:
                n_duplicate += 1
            _hash.append(o)

        if n_duplicate > 0 and not skip_check:
            lgg.warning('Duplicate experience: %d' % (n_duplicate))
            ask_sure_exit('Continue [y/n]?')
        self._hash = _hash

    def remake(self, indexs):
        ''' Update the curent tensors by selecting the ${indexs} '''

        self._lod = [self._lod[i] for i in indexs]
        self._tensors = []

        new_tensor = ExpTensor()
        consume_expe = 0
        self._tensors.append(new_tensor)
        while consume_expe < len(self._lod):
            d = self._lod[consume_expe]
            res = new_tensor.push_dict(d)
            if res is False:
                new_tensor = ExpTensor()
                self._tensors.append(new_tensor)
            else:
                consume_expe += 1
                new_tensor._size += 1

    def get_gt(self):
        ''' get Global Tensors.
            No _binding here...
        '''
        gt = {}
        for tensor in self._tensors:
            for k, v in tensor.items():
                _v = gt.get(k,[])
                gt[k] = _v + v
        return gt

    def get_keys(self):
       return list(self.get_gt())

    def table(self):
        tables = []
        for id, group in enumerate(self._tensors):
            src = self._ds[id].__name__
            spec = group.get('_name_expe', ['void'])[0]
            h = '=== %s > %s > %s expe ===' % (src, spec, group.get_size())
            tables.append(h)
            if self._bind:
                extra = [('_bind', self._bind[id])]

            if id == 0:
                headers = ['Params','Values']
            else:
                headers = ''

            tables.append(tabulate(extra+sorted(group.items(), key=lambda x:x[0]), headers=headers))

        return '\n'.join(tables)



class ExpDesign(dict, BaseObject):
    ''' An Ensemble composed of ExpTensors and ExpVectors.
        It is the base class to write group of specification.

        NOTES
        -----
        Special attribute meaning:
            _alias : dict
                use when self._name is called to translate keywords
    '''

    def __init__(self,  *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        # Not a Ultimate solution to keep a flexibility when defining Exp Design
        for k in dir(self):
            #_spec = ExpDesign((k, getattr(Netw, k)) for k in dir(Netw) if not k.startswith('__') )
            if not k.startswith('_'):
                v = getattr(self, k)
                #if not hasattr(v, '__call__'): # print a warning because hasattr call getattr in expSpace.
                if not callable(v): #  python >3.2
                    self[k] = v
        # @debug: add callable in reserved keyword
        self._reserved_keywords = list(set([w for w in dir(self) if w.startswith('_')] + ['_reserved_keywords']+dir(dict)+dir(BaseObject)))

        BaseObject.__init__(self)

    def _specs(self):
        return [ k for k  in self.keys() if k not in self._reserved_keywords ]


    @classmethod
    def _name(cls, l):
        if getattr(cls, '_alias', None):
            _alias =  cls._alias
        else:
            return l


        if isinstance(l, (set, list, tuple)):
            return [ _alias.get(i, i) for i in l ]
        elif isinstance(l, (dict, ExpSpace)):
            d = dict(l)
            for k, v in d.items():
                if isinstance(v, basestring) and v in _alias:
                    d[k] = _alias[v]
            return d
        else :
            return _alias.get(l, l)




# -*- coding: utf-8 -*-

import numpy, scipy;
import scipy.special;

"""
@author: Ke Zhai (zhaike@cs.umd.edu)

This is a python implementation of vanilla Indian Buffet process, based on variational inference.

References:
[1] Finale Doshi-Velez, Kurt T. Miller, Jurgen Van Gael, and Yee Whye Teh, Variational Inference for the Indian Buffet Process, Artificial Intelligence and Statistics (AISTATS), 2009.
"""
class VariationalBayes(object):
    """
    """
    def __init__(self,
                 snapshot_interval=10,
                 finite_mode=True,
                 model_likelihood_threshold=0.00001,
                 global_maximum_iteration=100):
        self._global_maximum_iteration = global_maximum_iteration;
        self._finite_mode = finite_mode;
        self._model_likelihood_threshold = model_likelihood_threshold;

        self._snapshot_interval = snapshot_interval;

    """
    @param data: a N-by-D matrix, representing N images, each of which contains D pixel
    @param truncation_level: truncation level for variational inference indian buffet process
    @param alpha: hyper-parameter defining the indian buffet process
    @param sigma_a: feature sigma
    @param sigma_x: data sigma
    """
    def _initialize(self, data, truncation_level=5, alpha=1., sigma_a=1., sigma_x=1.):
        self._X = data;
        (self._N, self._D) = self._X.shape;

        self._K = truncation_level;

        self._alpha = alpha;
        self._sigma_a = sigma_a;
        self._sigma_x = sigma_x;

        # tau
        self._tau = numpy.ones((2, self._K));
        if self._finite_mode:
            self._tau[0, :] = self._alpha / self._K;
            self._tau += 0.5 * numpy.min(1., self._alpha / self._K) * (numpy.random.random(self._tau.shape) - 0.5);
        else:
            self._tau[0, :] = self._alpha;
            self._tau += 0.5 * numpy.min(1., self._alpha) * (numpy.random.random(self._tau.shape) - 0.5);
        assert(self._tau.shape == (2, self._K));

        # nu
        self._nu = numpy.random.random((self._N, self._K));
        assert(self._nu.shape == (self._N, self._K));

        # phi
        self._phi_mean = numpy.random.normal(0., 1., (self._K, self._D)) * 0.01;
        self._phi_cov = numpy.random.normal(0., 1., (self._K, self._D)) ** 2 * 0.1;
        assert(self._phi_mean.shape == (self._K, self._D));
        assert(self._phi_cov.shape == (self._K, self._D));

    """
    """
    def update_phi(self):
        nu_sum_n = numpy.sum(self._nu, axis=0);
        assert(len(nu_sum_n) == self._K);
        nu_sum_k = numpy.sum(self._nu, axis=1);
        assert(len(nu_sum_k) == self._N);
        for k in numpy.random.permutation(xrange(self._K)):
            # we assume the covariance matrix is a diagonal matrix and we only store the diagonal terms
            self._phi_cov[k, :] = 1. / (1. / (self._sigma_a * self._sigma_a) + nu_sum_n[k] / (self._sigma_x * self._sigma_x));

            tmp_nu = numpy.delete(self._nu, k, 1);
            assert(tmp_nu.shape == (self._N, self._K - 1));
            tmp_phi_mean = numpy.delete(self._phi_mean, k, 0);
            assert(tmp_phi_mean.shape == (self._K - 1, self._D));
            phi_mean = numpy.dot(self._X.transpose(), self._nu[:, k][:, numpy.newaxis]).transpose();
            for n in xrange(self._N):
                phi_mean -= self._nu[n, k] * numpy.sum(tmp_nu[n, :] * tmp_phi_mean.transpose(), axis=1)[:, numpy.newaxis].transpose();
            assert(phi_mean.shape == (1, self._D));
            phi_mean /= self._sigma_x * self._sigma_x;
            self._phi_mean[k, :] = phi_mean * self._phi_cov[k, :];

        return

    """
    """
    def update_nu(self):
        var_theta_constant = self.compute_var_theta_constant();
        for k in numpy.random.permutation(xrange(self._K)):
            tmp_nu = numpy.delete(self._nu, k, 1);
            assert(tmp_nu.shape == (self._N, self._K - 1));
            tmp_phi_mean = numpy.delete(self._phi_mean, k, 0);
            assert(tmp_phi_mean.shape == (self._K - 1, self._D));

            for n in numpy.random.permutation(xrange(self._N)):
                x_nu_phi = self._X[n, :][numpy.newaxis, :] - numpy.sum(tmp_nu[n, :] * tmp_phi_mean.transpose(), axis=1)[:, numpy.newaxis].transpose();
                assert(x_nu_phi.shape == (1, self._D));
                var_theta = var_theta_constant[k] + 1. / (self._sigma_x * self._sigma_x) * numpy.dot(self._phi_mean[k, :][numpy.newaxis, :], x_nu_phi.transpose());
                self._nu[n, k] = 1. / (1. + numpy.exp(-var_theta));

        return;

    """
    """
    def update_tau(self):
        sum_nu = numpy.sum(self._nu, axis=0);
        assert(len(sum_nu) == self._K);
        if self._finite_mode:
            assert(len(sum_nu) == self._K);
            self._tau[0, :] = self._alpha / self._K + sum_nu;
            self._tau[1, :] = 1. + self._N - sum_nu;
        else:
            N_minus_sum_nu = self._N - sum_nu;
            for k in xrange(self._K):
                psi_tau = scipy.special.psi(self._tau);
                assert(psi_tau.shape == (2, self._K));

                psi_sum_tau = scipy.special.psi(numpy.sum(self._tau, axis=0));
                assert(len(psi_sum_tau) == self._K);

                psi_tau0_cumsum = numpy.hstack([0, numpy.cumsum(psi_tau[0, :-1])]);
                assert(len(psi_tau0_cumsum) == self._K);

                psi_sum_cumsum = numpy.cumsum(psi_sum_tau);
                assert(len(psi_sum_cumsum) == self._K);

                exponent = psi_tau[1, :] + psi_tau0_cumsum - psi_sum_cumsum;
                unnormalized = numpy.exp(exponent - numpy.max(exponent));
                assert(len(unnormalized) == self._K);

                qs = numpy.zeros((self._K, self._K));
                for m in xrange(k, self._K):
                    qs[m, 0:m + 1] = unnormalized[0:m + 1] / numpy.sum(unnormalized[0:m + 1]);

                self._tau[0, k] = numpy.sum(sum_nu[k:self._K]) + numpy.dot(N_minus_sum_nu[k + 1:self._K], numpy.sum(qs[k + 1:self._K, k + 1:self._K], axis=1)) + self._alpha;
                self._tau[1, k] = numpy.dot(N_minus_sum_nu[k:self._K], qs[k:self._K, k]) + 1;
        return;

    """
    """
    def compute_var_theta_constant(self):
        var_theta_constant = -0.5 / (self._sigma_x * self._sigma_x) * ((numpy.sum(self._phi_cov, axis=1) + numpy.sum(self._phi_mean * self._phi_mean, axis=1)));
        assert(len(var_theta_constant) == self._K);
        if self._finite_mode:
            var_theta_constant += scipy.special.psi(self._tau[0, :]) - scipy.special.psi(self._tau[1, :]);
        else:
            for k in xrange(self._K):
                var_theta_constant[k] += numpy.sum(scipy.special.psi(self._tau[0, 0:k]) - scipy.special.psi(numpy.sum(self._tau[:, 0:k])));
                var_theta_constant[k] -= self.compute_expected_pzk0_qjensen(k);
        return var_theta_constant;

    """
    """
    def compute_expected_pzk0_qjensen(self, k):
        assert(k >= 0 and k < self._K);
        tau = self._tau[:, 0:k + 1];
        assert(tau.shape == (2, k + 1));

        psi_tau = scipy.special.psi(tau);
        assert(psi_tau.shape == (2, k + 1));

        psi_sum_tau = scipy.special.psi(numpy.sum(tau, axis=0));
        assert(len(psi_sum_tau) == k + 1);

        psi_tau0_cumsum = numpy.hstack([0, numpy.cumsum(psi_tau[0, :-1])]);
        assert(len(psi_tau0_cumsum) == k + 1);

        psi_sum_cumsum = numpy.cumsum(psi_sum_tau);
        assert(len(psi_sum_cumsum) == k + 1);

        tmp = psi_tau[1, :] + psi_tau0_cumsum - psi_sum_cumsum;
        assert(len(tmp) == k + 1);

        q = numpy.exp(tmp - numpy.max(tmp));
        assert(len(q) == k + 1);

        q = q / numpy.sum(q);
        assert(len(q) == k + 1);

        # compute the lower bound
        lower_bound = numpy.sum(q * (tmp - numpy.log(q)));

        return lower_bound

    """
    """
    def learning(self, iteration=0, directory="../../output/tmp-output"):
        if iteration <= 0:
            iteration = self._global_maximum_iteration;

        print( "likelihood is", self.velb(), endline='')

        for i in xrange(iteration):
            self.update_tau();
            self.update_phi();
            self.update_nu();
            print( "likelihood is", self.velb())

        print( "learning finished...")

    """
    compute the variational evidence lower bound (i.e. velb) in log scale
    """
    def velb(self):
        log_likelihood = numpy.zeros(5);

        psi_tau = scipy.special.psi(self._tau);
        assert(psi_tau.shape == (2, self._K));
        psi_sum_tau = scipy.special.psi(numpy.sum(self._tau, axis=0)[numpy.newaxis, :]);
        assert(psi_sum_tau.shape == (1, self._K));

        if self._finite_mode:
            # compute the probability of feature
            log_likelihood[0] = self._K * numpy.log(self._alpha / self._K) + (self._alpha / self._K - 1.) * numpy.sum(psi_tau[0, :] - psi_sum_tau);
            # compute the probability of feature statistics
            log_likelihood[1] = numpy.sum(self._nu * psi_tau[0, :]) + numpy.sum((1. - self._nu) * psi_tau[1, :]) - self._N * numpy.sum(psi_sum_tau);
        else:
            # compute the probability of feature
            log_likelihood[0] = self._K * numpy.log(self._alpha) + (self._alpha - 1.) * numpy.sum(psi_tau[0, :] - psi_sum_tau);
            # compute the probability of feature statistics
            for k in xrange(self._K):
                log_likelihood[1] += numpy.sum(self._nu[:, k]) * numpy.sum(psi_tau[0, :k + 1] - psi_sum_tau[0, :k + 1]);
                log_likelihood[1] += numpy.dot((self._N - numpy.sum(self._nu[:, k])), self.compute_expected_pzk0_qjensen(k));

        # compute the probability of feature distribution
        log_likelihood[2] = -0.5 * self._K * self._D * numpy.log(2 * numpy.pi * self._sigma_a * self._sigma_a);
        log_likelihood[2] -= 0.5 / (self._sigma_a * self._sigma_a) * (numpy.sum(self._phi_cov) + numpy.sum(self._phi_mean * self._phi_mean));

        # compute the probability of data likelihood
        tmp_log_likelihood = numpy.sum(self._X * self._X) - 2 * numpy.sum(self._nu * numpy.dot(self._X, self._phi_mean.transpose()));
        tmp_1 = numpy.dot(numpy.ones((self._N, self._D)), (self._phi_cov + self._phi_mean ** 2).transpose());
        tmp_log_likelihood += numpy.sum(self._nu * tmp_1);
        tmp_1 = numpy.dot(self._nu, self._phi_mean);
        tmp_2 = numpy.sum(numpy.dot(self._nu ** 2, self._phi_mean ** 2));
        tmp_log_likelihood += numpy.sum(tmp_1 * tmp_1) - numpy.sum(tmp_2);

        log_likelihood[3] = -0.5 * self._N * self._D * numpy.log(2 * numpy.pi * self._sigma_x * self._sigma_x);
        log_likelihood[3] -= 0.5 / (self._sigma_x * self._sigma_x) * tmp_log_likelihood

        # entropy of the proposed distribution
        lngamma_tau = scipy.special.gammaln(self._tau);
        assert(lngamma_tau.shape == (2, self._K));
        lngamma_sum_tau = scipy.special.gammaln(numpy.sum(self._tau, axis=0)[numpy.newaxis, :]);
        assert(lngamma_sum_tau.shape == (1, self._K));

        # compute the entropy of the distribution
        log_likelihood[4] = numpy.sum(lngamma_tau[0, :] + lngamma_tau[1, :] - lngamma_sum_tau);
        log_likelihood[4] -= numpy.sum((self._tau[0, :] - 1) * psi_tau[0, :] + (self._tau[1, :] - 1) * psi_tau[1, :]);
        log_likelihood[4] += numpy.sum((self._tau[0, :] + self._tau[1, :] - 2) * psi_sum_tau);

        assert(numpy.all(self._phi_cov > 0));
        assert(numpy.all(self._nu >= 0) and numpy.all(self._nu <= 1));
        log_likelihood[4] += 0.5 * self._K * self._D * numpy.log(2 * numpy.pi * numpy.e);
        log_likelihood[4] += 0.5 * numpy.sum(numpy.log(self._phi_cov));
        #log_likelihood[4] += 0.5 * numpy.log(numpy.sqrt(numpy.sum(self._phi_cov * self._phi_cov, axis=1)));
        log_likelihood[4] -= numpy.sum(self._nu * numpy.log(self._nu) + (1. - self._nu) * numpy.log(1. - self._nu));

        return numpy.sum(log_likelihood);

if __name__ == "__main__":
    import scipy.io;
    import numpy;

    # load the data from the matrix
    mat_vals = scipy.io.loadmat('../../data/cambridge-bars/block_image_set.mat');
    true_weights = mat_vals['trueWeights']
    features = mat_vals['features']
    data = mat_vals['data']

    # initialize the model
    ibp = VariationalBayes(10, False, 0.00001, 100)
    #ibp._initialize(data[1:10, :], 4, true_weights, 1., 1., 1.);
    ibp._initialize(data[1:10, :])
    ibp.learning(10);

    # If matplotlib is installed, plot ground truth vs learned factors
    import matplotlib.pyplot as P
    from util.scaled_image import scaledimage

    # Intensity plots of
    # -ground truth factor-feature weights (top)
    # -learned factor-feature weights (bottom)
    K = max(len(true_weights), len(ibp._phi_mean))
    (fig, subaxes) = P.subplots(2, K)
    for sa in subaxes.flatten():
        sa.set_visible(False)
    fig.suptitle('Ground truth (top) vs learned factors (bottom)')
    for (idx, trueFactor) in enumerate(true_weights):
        ax = subaxes[0, idx]
        ax.set_visible(True)
        scaledimage(trueFactor.reshape(6, 6),
                    pixwidth=3, ax=ax)
    for (idx, learnedFactor) in enumerate(ibp._phi_mean):
        ax = subaxes[1, idx]
        scaledimage(learnedFactor.reshape(6, 6),
                    pixwidth=3, ax=ax)
        ax.set_visible(True)
    P.show()



# -*- coding: utf-8 -*-


# **kwargs !?
def lda_gensim(self, data=None, data_t=None, id2word=None, save=False, model='ldamodel', load=False, updatetype='batch'):

    try:
        sys.path.insert(1, '../../gensim')
        import gensim as gsm
        from gensim.models import ldamodel, ldafullbaye
        Models = {'ldamodel': ldamodel, 'ldafullbaye': ldafullbaye, 'hdp': 1}
    except:
        pass

    fname = self.output_path if self.write else None
    delta = self.hyperparams['delta']
    alpha = 'asymmetric'
    K = self.expe['K']

    data = kwargs['data']
    data_t = kwargs.get('data_t')

    if load:
        return Models[model].LdaModel.load(fname)

    if hasattr(data, 'tocsc'):
        # is csr sparse matrix
        data = data.tocsc()
        data = gsm.matutils.Sparse2Corpus(data, documents_columns=False)
        if heldout_data is not None:
            heldout_data = heldout_data.tocsc()
            heldout_data = gsm.matutils.Sparse2Corpus(heldout_data, documents_columns=False)
    elif isanparray:
        # up tocsc ??!!! no !
        dense2corpus
    # Passes is the iterations for batch onlines and
    # iteration the max it in the gamma treshold test
    # loop Batch setting !
    if updatetype == 'batch':
        lda = Models[model].LdaModel(data, id2word=id2word, num_topics=K, alpha=alpha, eta=delta,
                                     iterations=100, eval_every=None, update_every=None, passes=self.iterations, chunksize=200000,
                                     fname=fname, heldout_corpus=heldout_data)
    elif updatetype == 'online':
        lda = Models[model].LdaModel(data, id2word=id2word, num_topics=K, alpha=alpha, eta=delta,
                                     iterations=100, eval_every=None, update_every=1, passes=1, chunksize=2000,
                                     fname=fname, heldout_corpus=heldout_data)

    if save:
        lda.expElogbeta = None
        lda.sstats = None
        lda.save(fname)
    return lda




import numpy as np
import scipy as sp

from scipy.io.matlab import loadmat
from scipy.sparse import lil_matrix
try:
    from rescal import rescal_als
except ImportError as e:
    print('Import Error: %s' % e)


from .modelbase import ModelBase

class Rescal_als(ModelBase):

    def _init_params(self, frontend):
        frontend = self.frontend

        # Save the testdata
        self.data_test = frontend.data_test_w

        # For fast computation of bernoulli pmf.
        self._w_a = self.data_test[:,2].T.astype(int)
        self._w_a[self._w_a > 0] = 1
        self._w_a[self._w_a == 0] = -1
        self._w_b = np.zeros(self._w_a.shape, dtype=int)
        self._w_b[self._w_a == -1] = 1

        self._K = self.expe.K

    def _reduce_latent(self):
        return self._theta, self._phi

    def likelihood(self, theta=None, phi=None):
        if theta is None:
            theta = self._theta
        if phi is None:
            phi = self._phi

        qijs = []
        for i,j, xij in self.data_test:
            qijs.append( theta[i].dot(phi[0]).dot(theta[j]) )

        qijs = np.array(qijs)
        likelihood = 1 / (1 + np.exp(-qijs))

        return likelihood

    def compute_entropy(self, theta=None, phi=None, **kws):
        if 'likelihood' in kws:
            pij = kws['likelihood']
        else:
            if theta is None:
                theta, phi = self._reduce_latent()
            pij = self.likelihood(theta, phi)

        ll = pij * self._w_a + self._w_b

        ll[ll<=1e-300] = 1e-300
        # Log-likelihood
        ll = np.log(ll).sum()
        # Perplexity is 2**H(X).
        return ll

    def compute_roc(self, theta=None, phi=None, **kws):
        from sklearn.metrics import roc_curve, auc, precision_recall_curve

        if 'likelihood' in kws:
            pij = kws['likelihood']
        else:
            if theta is None:
                theta, phi = self._reduce_latent()
            pij = self.likelihood(theta, phi)

        weights = np.squeeze(self.data_test[:,2].T)

        y_true = weights.astype(bool)*1
        self._probas = pij
        self._y_true = y_true

        fpr, tpr, thresholds = roc_curve(y_true, pij)
        roc = auc(fpr, tpr)
        return roc

    def compute_pr(self, *args, **kwargs):
        from sklearn.metrics import average_precision_score
        return average_precision_score(self._y_true, self._probas)

    def compute_wsim(self, *args, **kws):
        return None

    def generate(self, N=None, K=None, hyperparams=None, mode='predictive', symmetric=True, **kwargs):
        likelihood = self.likelihood()
        #likelihood[likelihood <= 0.5] = 0
        #likelihood[likelihood > 0.5] = 1
        #Y = likelihood
        Y = sp.stats.bernoulli.rvs(likelihood)
        return Y

    def fit(self, frontend):
        self._init(frontend)
        K = self.expe.K
        y = frontend.adj()
        data = [y]

        self.log.info("Fitting `%s' model" % (type(self)))
        A, R, fit, itr, exectimes = rescal_als(data, K, init='nvecs', lambda_A=10, lambda_R=10)

        self._theta = A
        self._phi = R

        self.log.info('rescal fit info: %s; itr: %s, exectimes: %s' % (fit, itr, exectimes))

        self.compute_measures()
        if self.expe.get('_write'):
            self.write_current_state(self)






def rescal(X, K):

    ## Set logging to INFO to see RESCAL information
    #logging.basicConfig(level=logging.INFO)

    ## Load Matlab data and convert it to dense tensor format
    #T = loadmat('data/alyawarra.mat')['Rs']
    #X = [lil_matrix(T[:, :, k]) for k in range(T.shape[2])]

    X = [sp.sparse.csr_matrix(X)]
    A, R, fit, itr, exectimes = rescal_als(X, K, init='nvecs', lambda_A=10, lambda_R=10)

    theta =  A.dot(R).dot(A.T)
    Y = 1 / (1 + np.exp(-theta))
    Y =  Y[:,0,:]
    Y[Y <= 0.5] = 0
    Y[Y > 0.5] = 1
    #Y = sp.stats.bernoulli.rvs(Y)
    return Y




# -*- coding: utf-8 -*-

import collections
import logging
lgg = logging.getLogger('root')

import numpy as np
import scipy as sp

from scipy.special import digamma
from numpy.random import dirichlet, gamma, poisson, binomial, beta

from pymake.frontend.frontend import DataBase
from ml.model.modelbase import GibbsSampler
from .hdp import MSampler, BetaSampler

from pymake.util.math import lognormalize, categorical


# Implementation of Teh et al. (2005) Gibbs sampler for the Hierarchical Dirichlet Process: Direct assignement.

#import warnings
#warnings.simplefilter('error', RuntimeWarning)

""" @Todo
* Control the calcul of perplexity at each iterations. Costly !

* HDP direct assignement:
Network implementation (2*N restaurant, symmetric, p(k,l|.) = p(k)p(l) etc)

"""

# Broadcast class based on numpy matrix
# Assume delta a scalar.
class Likelihood(object):

    def __init__(self, delta, data, **config):
        if type(data) is np.ndarray:
            # Error when streaming ? sppy ?
            #self.data_mat = sp.sparse.csr_matrix(data)
            self.data_mat = data
        elif data.format == 'csr':
            self.data_mat = data
        else:
            raise NotImplementedError('type %s unknow as corpus' % type(data))
        self.data = DataBase.sparse2stream(self.data_mat)
        assert(len(self.data) > 1)
        assert(type(delta) in (int, float))
        self.delta = delta

        self.data_dims = self.get_data_dims()
        self.nnz = self.get_nnz()
        # Vocabulary size
        self.nfeat = self.get_nfeat()

        # Cst for CGS of DM and scala delta as prior.
        self.delta = delta if isinstance(delta, np.ndarray) else np.asarray([delta] * self.nfeat)
        self.w_delta = self.delta.sum()

        print(self.data_mat.shape, self.nfeat)
        assert(self.data_mat.shape[1] == self.nfeat)
        #self.data_mat = sppy.csarray(self.data_mat)

    def compute(self, j, i, k_ji):
        return self.loglikelihood(j, i, k_ji)

    def get_nfeat(self):
        return np.vectorize(max)(self.data).max() + 1

    def get_data_dims(self):
        data_dims = np.vectorize(len)(self.data)
        return list(data_dims)

    def get_nnz(self):
        return sum(self.data_dims)

    def words_k(self, where_k):
        for group in range(len(self.data)):
            word_gen = self.data[group][where_k[group]]
            for word in word_gen:
                yield word

    def make_word_topic_counts(self, z, K):
        word_topic_counts = np.zeros((self.nfeat, K), dtype=int)

        for k in range(K):
            where_k =  np.array([np.where(zj == k)[0] for zj in z])
            words_k_dict = collections.Counter(self.words_k(where_k))
            word_topic_counts[words_k_dict.keys(), k] = words_k_dict.values()

        self.word_topic_counts = word_topic_counts

    def loglikelihood(self, j, i, k_ji):
        w_ji = self.data[j][i]
        self.word_topic_counts[w_ji, k_ji] -= 1
        self.total_w_k[k_ji] -= 1
        log_smooth_count_ji = np.log(self.word_topic_counts[w_ji] + self.delta[w_ji])

        return log_smooth_count_ji - np.log(self.total_w_k + self.w_delta)

class ZSampler(object):
    # Alternative is to keep the two count matrix and
    # The docment-word topic array to trace get x(-ij) topic assignment:
        # C(word-topic)[w, k] === n_dotkw
        # C(document-topic[j, k] === n_jdotk
        # z DxN_k topic assignment matrix
        # --- From this reconstruct theta and phi

    def __init__(self, alpha_0, likelihood, K_init=0, data_t=None):
        self.K_init = K_init
        self.alpha_0 = alpha_0
        self.likelihood = likelihood
        self.data_dims = likelihood.data_dims
        self.J = len(self.data_dims)
        self.z = self._init_topics_assignement()
        self.doc_topic_counts = self.make_doc_topic_counts()
        if not hasattr(self, 'K'):
            # Nonparametric Case
            self.purge_empty_topics()
        self.likelihood.make_word_topic_counts(self.z, self.get_K())
        self.likelihood.total_w_k = self.likelihood.word_topic_counts.sum(0)

        if data_t is not None:
            self.data_t = data_t
            self.data_t_w = DataBase.sparse2stream(data_t)
            self.z_t = self._init_topics_assignement(data_t)
            self.doc_topic_counts_t = self.make_doc_topic_counts(self.data_t)

        # if a tracking of topics indexis pursuit,
        # pay attention to the topic added among those purged...(A topic cannot be added and purged in the same sample iteration !)
        self.last_purged_topics = []

    # @data_dims: list of number of observations per document/instance
    def _init_topics_assignement(self, data=None):
        if data is None:
            data_dims = self.data_dims
        else:
            data_dims = self.data_t.sum(1)
        alpha_0 = self.alpha_0

        # Poisson way
        #z = np.array( [poisson(alpha_0, size=dim) for dim in data_dims] )

        # Random way
        K = self.K_init
        z = np.array( [np.random.randint(0, K, (dim)) for dim in data_dims] )

        # LDA way
        # improve local optima ?
        #todo ?

        return z

    def sample(self):
        # Add pnew container
        self._update_log_alpha_beta()
        self.doc_topic_counts =  np.column_stack((self.doc_topic_counts, np.zeros(self.J, dtype=int)))
        self.likelihood.word_topic_counts =  np.column_stack((self.likelihood.word_topic_counts, np.zeros(self.likelihood.nfeat, dtype=int)))
        self.likelihood.total_w_k = self.likelihood.word_topic_counts.sum(0)

        lgg.info( 'Sample z...')
        lgg.debug( '#Doc \t nnz \t  #topic')
        doc_order = np.random.permutation(self.J)
        for doc_iter, j in enumerate(doc_order):
            nnz =  self.data_dims[j]
            lgg.debug( '%d \t %d \t %d' % ( doc_iter , nnz, self.doc_topic_counts.shape[1]-1 ))
            nnz_order = np.random.permutation(nnz)
            for i in nnz_order:
                params = self.prob_zji(j, i, self._K + 1)
                sample_topic = categorical(params)
                self.z[j][i] = sample_topic

                # Regularize matrices for new topic sampled
                if sample_topic == self.doc_topic_counts.shape[1] - 1:
                    self._K += 1
                    #print 'Simplex probabilities: %s' % (params)
                    col_doc = np.zeros((self.J, 1), dtype=int)
                    col_word = np.zeros((self.likelihood.nfeat, 1), dtype=int)
                    self.doc_topic_counts = np.hstack((self.doc_topic_counts, col_doc))
                    self.likelihood.word_topic_counts = np.hstack((self.likelihood.word_topic_counts, col_word))
                    self.likelihood.total_w_k = self.likelihood.word_topic_counts.sum(0)

                # Update count matrixes
                self.doc_topic_counts[j, sample_topic] += 1
                self.likelihood.word_topic_counts[self.likelihood.data[j][i], sample_topic] += 1
                self.likelihood.total_w_k[sample_topic] += 1

        # Remove pnew container
        self.doc_topic_counts = self.doc_topic_counts[:, :-1]
        self.likelihood.word_topic_counts = self.likelihood.word_topic_counts[:,:-1]
        self.purge_empty_topics()

        return self.z

    def sample_heldout(self):
        J = self.data_t.shape[0]
        K = self.get_K()

        # Manage Alpha prior
        if not hasattr(self, 'logalpha'):
            alpha = np.exp(self.log_alpha_beta[:-1])
        else:
            alpha = np.exp(self.logalpha)

        doc_order = np.random.permutation(J)
        for doc_iter, j in enumerate(doc_order):
            nnz =  self.data_t[j].sum()
            lgg.debug( '%d \t %d \t %d' % ( doc_iter , nnz, K ))
            nnz_order = np.random.permutation(nnz)
            for i in nnz_order:

                k_ji = self.z_t[j][i]
                self.doc_topic_counts_t[j, k_ji] -=1

                params = np.log(self.doc_topic_counts_t[j] + alpha) + np.log(self._phi[self.data_t_w[j][i], k_ji])
                params =  lognormalize(params[:K])

                sample_topic = categorical(params)
                self.z_t[j][i] = sample_topic

                self.doc_topic_counts_t[j, sample_topic] += 1

        return self.z_t

    def make_doc_topic_counts(self, data=None):
        if data is None:
            J = self.J
            z = self.z
        else:
            J = data.shape[0]
            z = self.z_t
        K = self.get_K()
        counts = np.zeros((J, K), dtype=int)
        for j, d in enumerate(z):
            counts[j] = np.bincount(d, minlength=K)

        return counts

    def _update_log_alpha_beta(self):
        self.log_alpha_beta = np.log(self.alpha_0) + np.log(self.betasampler.beta)

    # Remove empty topic in nonparametric case
    def purge_empty_topics(self):
        counts = self.doc_topic_counts

        dummy_topics = []
        # Find empty topics
        for k, c in enumerate(counts.T):
            if c.sum() == 0:
                dummy_topics.append(k)
        for k in sorted(dummy_topics, reverse=True):
            counts = np.delete(counts, k, axis=1)
            self._K -= 1
            if hasattr(self.likelihood, 'word_topic_counts'):
                self.likelihood.word_topic_counts = np.delete(self.likelihood.word_topic_counts, k, axis=1)
            # Regularize Z
            for d in self.z:
                d[d > k] -= 1
            # Regularize alpha_beta, minus one the pnew topic
            if hasattr(self, 'log_alpha_beta') and k < len(self.log_alpha_beta)-1:
                self.log_alpha_beta = np.delete(self.log_alpha_beta, k)
                self.betasampler.beta = np.delete(self.betasampler.beta, k)

        self.last_purged_topics = dummy_topics
        if len(dummy_topics) > 0:
            lgg.info( 'zsampler: %d topics purged' % (len(dummy_topics)))
        self.doc_topic_counts =  counts

    def add_beta_sampler(self, betasampler):
        self.betasampler = betasampler
        self._update_log_alpha_beta()

    def get_K(self):
        if not hasattr(self, '_K'):
            self._K =  np.max(np.vectorize(np.max)(self.z)) + 1
        return self._K

    # Compute probabilityy to sample z_ij = k for each [K].
    # K is would be fix or +1 for nonparametric case
    def prob_zji(self, j, i, K):
        k_ji = self.z[j][i]
        self.doc_topic_counts[j, k_ji] -=1

        # Manage Alpha prior
        if not hasattr(self, 'logalpha'):
            log_alpha_beta = self.log_alpha_beta
            new_k = K - len(log_alpha_beta)
            if new_k > 0:
                log_alpha_beta = np.hstack((log_alpha_beta, np.ones((new_k,))*log_alpha_beta[-1]))
            alpha = np.exp(log_alpha_beta)
        else:
            alpha = np.exp(self.logalpha)

        params = np.log(self.doc_topic_counts[j] + alpha) + self.likelihood.compute(j, i, k_ji)
        return lognormalize(params[:K])

    def get_log_alpha_beta(self, k):
        old_max = self.log_alpha_beta.shape[0]

        if k > (old_max - 1):
            return self.log_alpha_beta[old_max - 1]
        else:
            return self.log_alpha_beta[k]

    def clean(self):
        self.K = self.doc_topic_counts.shape[1]

    def predictive_topics(self):
        # check if perplecxy is equal if removing dummy empty topics...
        if not hasattr(self, 'logalpha'):
            alpha = np.exp(self.log_alpha_beta[:-1])
        else:
            alpha = np.exp(self.logalpha)
        delta = self.likelihood.delta
        K = len(alpha)
        J = self.data_t.shape[0]

        # Recontruct Documents-Topic matrix
        _theta = self.doc_topic_counts_t + np.tile(alpha, (J, 1))
        _theta = (_theta.T / _theta.sum(axis=1)).T
        return _theta

    def estimate_latent_variables(self):
        # check if perplecxy is equal if removing dummy empty topics...
        if not hasattr(self, 'logalpha'):
            log_alpha_beta = self.log_alpha_beta
            new_k = self.get_K()+1 - len(log_alpha_beta)
            if new_k > 0:
                gmma = log_alpha_beta[-1]
                log_alpha_beta = np.hstack((log_alpha_beta, np.ones((new_k,))*gmma))
            # Remove empty possibly new topic
            alpha = np.exp(log_alpha_beta[:-1])
        else:
            alpha = np.exp(self.logalpha)
        delta = self.likelihood.delta
        K = len(alpha)

        # Recontruct Documents-Topic matrix
        _theta = self.doc_topic_counts + np.tile(alpha, (self.J, 1))
        self._theta = (_theta.T / _theta.sum(axis=1)).T

        # Recontruct Words-Topic matrix
        _phi = self.likelihood.word_topic_counts.T + np.tile(delta, (K, 1))
        self._phi = (_phi.T / _phi.sum(axis=1))

        return self._theta, self._phi

    # Mean can be arithmetic or geometric
    def perplexity(self, data=None, mean='arithmetic'):
        phi = self._phi
        if data is None:
            data = self.likelihood.data_mat
            nnz = self.likelihood.nnz
            theta = self._theta
        else:
            data = self.data_t
            nnz = self.data_t.sum()
            theta = self.predictive_topics()

        ### based on aritmetic mean

        ### Loop Approach
        #entropy = 0.0
        #_indices = lambda x: x.nonzero()[1]
        #for j in range(self.J):
        #    data_j = [ (i, data[j, i]) for i in _indices(data[j]) ]
        #    entropy += np.sum( cnt_wi * np.log(theta[j] * phi[w_ji]).sum() for w_ji, cnt_wi in data_j )

        ### Vectorized approach
        # < 2s for kos and nips k=50, quite fast
        entropy = data.multiply( np.log( theta.dot(phi.T) )).sum()
        #entropy = (data.hadamard( sppy.csarray(np.log( theta.dot(phi.T) )) )).sum()

        perplexity = np.exp(-entropy / nnz)
        return perplexity

class ZSamplerParametric(ZSampler):
    # Parametric Version of HDP sampler. Number of topics fixed.

    def __init__(self, alpha_0, likelihood, K, alpha='asymmetric', data_t=None):
        self.K = self.K_init = self._K =  K
        if 'alpha' is 'symmetric':
            alpha = np.ones(K)*1/K
        elif 'alpha' == 'asymmetric':
            alpha = np.asarray([1.0 / (i + np.sqrt(K)) for i in range(K)])
            alpha /= alpha.sum()
        else:
            alpha = np.ones(K)*alpha_0
        self.logalpha = np.log(alpha)
        super(ZSamplerParametric, self).__init__(alpha_0, likelihood, self.K, data_t=data_t)

    def sample(self):
        print( 'Sample z...')
        lgg.debug( '#Doc \t #nnz\t #Topic')
        doc_order = np.random.permutation(self.J)
        for doc_iter, j in enumerate(doc_order):
            nnz =  self.data_dims[j]
            lgg.debug( '%d \t %d \t %d' % ( doc_iter , nnz, self.K ))
            nnz_order = np.random.permutation(nnz)
            for i in nnz_order:
                params = self.prob_zji(j, i, self.K)
                sample_topic = categorical(params)
                self.z[j][i] = sample_topic

                self.doc_topic_counts[j, sample_topic] += 1
                self.likelihood.word_topic_counts[self.likelihood.data[j][i], sample_topic] += 1
                self.likelihood.total_w_k[sample_topic] += 1

        return self.z

    def get_K(self):
        return self.K

    def get_log_alpha_beta(self, k):
        return self.logalpha[k]

    def clean(self):
        pass

class NP_CGS(object):

    # Joint Sampler of topic Assignement, table configuration, and beta proportion.
    # ref to direct assignement Sampling in HDP (Teh 2006)
    def __init__(self, zsampler, msampler, betasampler, hyper='auto'):
        zsampler.add_beta_sampler(betasampler)

        self.zsampler = zsampler
        self.msampler = msampler
        self.betasampler = betasampler

        msampler.sample()
        betasampler.sample()

        if hyper.startswith( 'auto' ):
            self.hyper = hyper
            self.a_alpha = 10
            self.b_alpha = 0.2
            self.a_gmma = 10
            self.b_gmma = 0.2
            self.optimize_hyper_hdp()
        elif hyper.startswith( 'fix' ):
            self.hyper = hyper
        else:
            raise NotImplementedError('Hyperparameters optimization ?')

    def optimize_hyper_hdp(self):
        # Optimize \alpha_0
        m_dot = self.msampler.m_dotk.sum()
        alpha_0 = self.zsampler.alpha_0
        n_jdot = np.array(self.zsampler.data_dims)
        #norm = np.linalg.norm(n_jdot/alpha_0)
        #u_j = binomial(1, n_jdot/(norm* alpha_0))
        u_j = binomial(1, n_jdot/(n_jdot + alpha_0))
        v_j = beta(alpha_0 + 1, n_jdot)
        new_alpha0 = gamma(self.a_alpha + m_dot - u_j.sum(), 1/( self.b_alpha - np.log(v_j).sum()), size=5).mean()
        self.zsampler.alpha_0 = new_alpha0

        # Optimize \gamma
        K = self.zsampler._K
        gmma = self.betasampler.gmma
        #norm = np.linalg.norm(m_dot/gmma)
        #u = binomial(1, m_dot / (norm*gmma))
        u = binomial(1, m_dot / (m_dot + gmma))
        v = beta(gmma + 1, m_dot)
        new_gmma = gamma(self.a_gmma + K -1 + u, 1/(self.b_gmma - np.log(v)), size=5).mean()
        self.betasampler.gmma = new_gmma

        print('alpha a, b: %s, %s ' % (self.a_alpha + m_dot - u_j.sum(), 1/( self.b_alpha - np.log(v_j).sum())))
        print( 'hyper sample: alpha_0: %s gamma: %s' % (new_alpha0, new_gmma))
        return

    def sample(self):
            z = self.zsampler.sample()
            m = self.msampler.sample()
            beta = self.betasampler.sample()

            if self.hyper.startswith('auto'):
                self.optimize_hyper_hdp()

            if hasattr(self.zsampler, 'data_t') and self.zsampler.data_t is not None:
                self.zsampler.sample_heldout()

            return z, m, beta

class CGS(object):

    def __init__(self, zsampler):
        self.zsampler = zsampler

    def sample(self):

        z = self.zsampler.sample()

        if hasattr(self.zsampler, 'data_t') and self.zsampler.data_t is not None:
            self.zsampler.sample_heldout()

        return z

class GibbsRun(GibbsSampler):
    __abstractmethods__ = 'model'
    def __init__(self, sampler,  data_t=None, **kwargs):

        self.burnin = kwargs.get('burnin',  0.05) # Ratio of iteration
        self.thinning = kwargs.get('thinning',  1)
        self.comm = dict() # Empty dict to store communities and blockmodel structure
        self.data_t = data_t
        self._csv_typo = 'it it_time entropy_train entropy_test K alpha gamma alpha_mean delta_mean alpha_var delta_var'
        self.fmt = '%d %.4f %.8f %.8f %d %.8f %.8f %.4f %.4f %.4f %.4f'
        #self.fmt = '%s %s %s %s %s %s %s %s %s %s %s'
        GibbsSampler.__init__(self, sampler, **kwargs)


    def predict(self):
        lgg.error('todo')
        pass




# -*- coding: utf-8 -*-

import logging
lgg = logging.getLogger('root')

from ml.model.ibp.ilfm_gs import IBPGibbsSampling

class ilfm_cgs(IBPGibbsSampling):
    def __init__(self, expe, frontend):

        hyperparams = expe.get('hyperparams', {})
        sigma_w = 1.
        sigma_w_hyper_parameter = None #(1., 1.)
        alpha = hyperparams.get('alpha',1)

        alpha_hyper_parameter = expe.hyper
        assortativity = expe.get('homo')

        #if '_cgs' in expe.model:
        #    metropolis_hastings_k_new = False
        metropolis_hastings_k_new = True

        if assortativity == 2:
            raise NotImplementedError('Warning !: Metropolis Hasting not implemented with matrix normal. Exiting....')

        super(ilfm_cgs, self).__init__(expe, frontend,
                                       assortativity,
                                       alpha_hyper_parameter,
                                       sigma_w_hyper_parameter,
                                       metropolis_hastings_k_new)

        self._initialize(frontend, alpha, sigma_w, KK=expe.K)
        self.update_hyper(hyperparams)

        lgg.warn('Warning: K is IBP initialized...')

        self.normalization_fun = lambda x : 1/(1 + np.exp(-x))

class lfm_cgs(ilfm_cgs):
    def __init__(self, expe, frontend):
        super().__init__(expe, frontend)

        self.log.warning(''' Random initializatin intead of ibp ? ''')
        self._metropolis_hastings_k_new = False






import os, shutil, logging
from pymake import get_pymake_settings
from pymake.util.utils import colored


import whoosh as ws
import whoosh.highlight
from whoosh.qparser import QueryParser, OrGroup, AndGroup


class TerminalFormatter(ws.highlight.Formatter):

    def format_token(self, text, token, replace=False):
        # Use the get_text function to get the text corresponding to the
        # token
        tokentext = ws.highlight.get_text(text, token, replace)

        # Return the text as you want it to appear in the highlighted
        # string
        return "%s" % colored(tokentext, 'bold')


class IndexManager(object):

    _SCHEMA   = {'model' : ws.fields.Schema(name      = ws.fields.ID(stored = True),
                                            surname   = ws.fields.ID(stored = True),
                                            module    = ws.fields.ID(stored = True),
                                            category  = ws.fields.KEYWORD(stored = True),
                                            content   = ws.fields.TEXT),

                 'script' : ws.fields.Schema(scriptname    = ws.fields.ID(stored = True),
                                             scriptsurname = ws.fields.ID(stored = True),
                                             module    = ws.fields.ID(stored = True),
                                             method    = ws.fields.KEYWORD(stored = True),
                                             signature = ws.fields.TEXT(stored = True),
                                             content   = ws.fields.TEXT),

                 'spec' : ws.fields.Schema(module_name  = ws.fields.ID(stored = True),
                                           script_name  = ws.fields.ID(stored = True),
                                           expe_name    = ws.fields.ID(stored = True),
                                           content      = ws.fields.TEXT),
                }

    log = logging.getLogger('root')

    def __init__(self, default_index='model'):
        self._DATA_PATH = os.path.join(get_pymake_settings('project_data'), '.pmk')

        self._index_basename = 'ir_index'
        self._default_index = default_index
        self._ix = {} # Index store by key

    def get_index_path(self, name=None):
        name = name or self._default_index
        return os.path.join(self._DATA_PATH, self._index_basename, name + '/')

    def clean_index(self, name=None, schema=None, **kwargs):
        ''' make the index `name\' according to its `schema\' '''
        name = name or self._default_index
        index_path = self.get_index_path(name)
        if os.path.exists(index_path):
            shutil.rmtree(index_path)
        os.makedirs(index_path)

        if name in self._SCHEMA:
            schema = self._SCHEMA[name]
        else:
            raise NotImplementedError('Dont know what to do, no schema defined ...?')

        self._ix[name] = ws.index.create_in(index_path, schema)

        return self._ix[name]

    def load_index(self, name=None):
        name = name or self._default_index
        index_path = self.get_index_path(name)
        return ws.index.open_dir(index_path)

    def get_index(self, name=None, donttryerror=False):
        name = name or self._default_index
        if name in self._ix:
            return self._ix[name]
        elif os.path.exists(self.get_index_path(name)):
            try:
               return  self.load_index(name)
            except Exception as e:
                if donttryerror:
                    raise e
                else:
                    self.log.warning('Indexing file corrupted, re-indexing...')
                    self.build_indexes()
                    return self.get_index(name, donttryerror=True)

        else:
            return self.clean_index(name)

    def get_writer(self, reset=False, online=None, index=None):
        index = index or self._default_index
        if reset:
            ix = self.clean_index(index)
        else:
            ix = self.get_index(index)

        if online:
            import whoosh.writing
            if online is True:
                online = {}
            period = online.get('period', 600)
            limit = online.get('limit', 2)
            return ws.writing.BufferedWriter(ix, period=period, limit=limit)
        else:
            return ix.writer()

    def get_reader(self, index=None):
        index = index or self._default_index
        ix = self.get_index(index)
        return ix.searcher()

    @classmethod
    def build_indexes(cls, index_name=None):
        ''' Update the system index '''
        idx = cls()

        if index_name is None:
            schemas = idx._SCHEMA
        else:
            schemas = [index_name]

        for name  in schemas:
            func = 'update_' + name + '_index'
            builder = getattr(idx, func)
            builder()


    def update_corpus_index(self):
        raise NotImplementedError

    def update_spec_index(self):
        ''' Update the schema of the Spec index '''
        from pymake.io import SpecLoader
        model = 'spec'
        self.log.info('Building %s index...' % model)
        Specs = SpecLoader.get_atoms()
        writer = self.clean_index(model).writer()
        for scriptname, _content in Specs.items():
            self.log.info('\tindexing %s' % (str(scriptname)+str(_content['_module'])))

            for expe in _content['exp']:

                content = ''
                writer.add_document(script_name = _content['script_name'],
                                    module_name = _content['module_name'],
                                    expe_name = expe,
                                    content = content)
        writer.commit()

    def update_script_index(self):
        ''' Update the schema of the Scripts index '''
        from pymake.io import ScriptsLoader
        model = 'script'
        self.log.info('Building %s index...' % model)
        Scripts = ScriptsLoader.get_atoms()
        writer = self.clean_index(model).writer()
        for scriptname, _content in Scripts.items():
            self.log.info('\tindexing %s' % (str(scriptname)+str(_content['_module'])))

            # Loop is context/model dependant
            for method in _content['methods']:

                content = ''
                writer.add_document(scriptname = _content['scriptname'],
                                    scriptsurname = _content['scriptsurname'],
                                    module = _content['module'],
                                    method = method,
                                    content = content)
        writer.commit()

    def update_model_index(self):
        ''' Update the schema of the Models index '''
        from pymake.io import ModelsLoader
        model = 'model'
        self.log.info('Building %s index...' % model)
        models = ModelsLoader.get_atoms()
        writer = self.clean_index(model).writer()
        for surname, module in models.items():
            self.log.info('\tindexing %s' % (str(surname)+str(module)))

            # Loop is context/model dependant
            topos = ' '.join(set(module.__module__.split('.')[1:]))
            content = ' '.join((surname, module.__name__, module.__module__))

            writer.add_document(surname = surname,
                                name = module.__name__,
                                category = topos,
                                module = module.__module__,
                                content = content)
        writer.commit()


    # @debug : online searcher
    def _search(self, query='',  field=None,  index=None, terms=False, limit=None):
        ''' query (exaxct mathch) search '''
        index = index or self._default_index
        ix = self.get_index(index)
        fieldin = field or 'content'

        qp = QueryParser(fieldin, ix.schema)
        qp.add_plugin(ws.qparser.SingleQuotePlugin())
        query = qp.parse(query, normalize=False)
        with ix.searcher() as searcher:
            if terms is True:
                results = searcher.search(query, terms=True, limit=limit).matched_terms()
            else:
                results = list(searcher.search(query, limit=limit).items())

        return results

    def search(self, query='',  field=None,  index=None, limit=None):
        ''' Text search '''
        index = index or self._default_index
        limit = None if limit == 'all' else limit
        ix = self.get_index(index)
        fieldin = field or 'content'

        fuzzy = '~' in query
        wildcard = '*' in query
        plusminus = '+' in query or '-' in query
        multifield = ':' in query
        boost = '^' in query

        qp = QueryParser(fieldin, ix.schema, group=OrGroup)

        # Check the difference between the both, not sure ?
        qp.add_plugin(ws.qparser.SingleQuotePlugin())
        qp.add_plugin(ws.qparser.PhrasePlugin())

        if wildcard:
            qp.add_plugin(ws.qparser.WildcardPlugin())
        if fuzzy:
            qp.add_plugin(ws.qparser.FuzzyTermPlugin())
        if plusminus:
            qp.add_plugin(ws.qparser.PlusMinusPlugin())
        if multifield:
            qp.add_plugin(ws.qparser.MultifieldPlugin([fieldin]))
        if boost:
            qp.add_plugin(ws.qparser.BoostPlugin())

        query = qp.parse(query)
        with ix.searcher() as searcher:
                results = searcher.search(query, limit=limit)

                results.fragmenter = ws.highlight.SentenceFragmenter(maxchars=200, charlimit=100042)
                #results.fragmenter = ws.highlight.ContextFragmenter(maxchars=200, surround=43)
                results.formatter = TerminalFormatter()

                self._last_total_match = len(results)
                for r in results:
                    yield r


    def getbydocid(self, docid, index=None):
        ''' return the a document's stored fields in the index from docid '''
        index = index or self._default_index
        ix = self.get_index(index)
        with ix.searcher() as searcher:
            doc = searcher.stored_fields(docid)
        return doc

    def getfirst(self, query='', field=None, index=None):
        query = "'" + query + "'"
        results = self._search(query, field, index, limit=1)

        if not results:
            return None
        else:
            return self.getbydocid(results[0][0])

    # not need to commit ?! /conflict forward...
    #def delete_by_term(self, term, field):
    #    writer.delete_by_term('hash', doc['hash'])

    def getbydocids(self, docids, index=None):
        ''' return the a list of document's stored fields in the index from docids '''
        index = index or self._default_index
        ix = self.get_index(index)
        docs = []
        with ix.searcher() as searcher:
            for docid in docids:
                docs.append(searcher.stored_fields(docid))
        return docs

    # @debug : online searcher
    # @debug : get a list of terms (mongo projection equivalent ?!)
    def query(self, field=None, index=None, terms=False, donttryerror=False):
        ''' return all object that have the field entry set '''
        index = index or self._default_index
        ix = self.get_index(index)
        field = field or ix.schema.stored_names()[0]

        query = ws.query.Every(field)
        try:
            with ix.searcher() as searcher:
                results = searcher.search(query, limit=None)
                if terms is False:
                    results = [r[field] for r in results]
                elif isinstance(terms, str):
                    results = dict((o[field], o[terms]) for o in results)
                else:
                    results = [dict(r) for r in results]
        except FileNotFoundError as e:
            if donttryerror:
                raise e
            else:
                self.log.warning('Indexing file corrupted, re-indexing...')
                self.build_indexes()

                return self.query(field, index, tems, donttryerror=True)


        return results


if __name__ == '__main__':
    ix = IndexManager()
    #ix.update_model_index()
    q = ix.query(field='surname')
    for r in q:
        print(r)



# -*- coding: utf-8 -*-

import numpy as np
import scipy as sp

from pymake.util.math import lognormalize, categorical, sorted_perm, adj_to_degree, gem
from ml.model.modelbase import GibbsSampler

#@network class frontend :
#   * pb of method conlicts, save, purge etc..fond a way
class immsb_cvb(GibbsSampler):

    # *args ?
    def _init_params(self):
        ### The time Limitations are @heere
        # frontend integration ?
        _len = {}
        _len['K'] = self.expe.get('K')
        _len['N'] = self.frontend.getN()
        _len['nfeat'] = self.frontend.get_nfeat()
        data_ma = self.frontend.data_ma
        _len['nnz'] = self.frontend.ma_nnz()
        _len['nnz_t'] = self.frontend.ma_nnz_t()
        _len['dims'] = self.frontend.ma_dims()
        _len['ones'] = (data_ma == 1).sum()
        _len['zeros'] = (data_ma == 0).sum()
        self._len = _len

        # Stream Parameters
        self.iterations = self.expe.get('iterations', 1)


        # Hyperparams
        delta = self.expe['hyperparams']['delta']
        self.hyper_phi = np.asarray(delta) if isinstance(delta, (np.ndarray, list, tuple)) else np.asarray([delta] * self._len['nfeat'])

        alpha = self.expe['hyperparams']['alpha'] # unused
        self.hyper_theta = np.asarray([1.0 / (i + np.sqrt(self._len['K'])) for i in range(self._len['K'])])
        self.hyper_theta /= self.hyper_theta.sum()

        self.hyper_phi_sum = self.hyper_phi.sum()
        self.hyper_theta_sum = self.hyper_theta.sum()

        # Sufficient Statistics
        self._ss = self._random_cvb_init()

        self.frontend._set_rawdata_for_likelihood_computation()


    def _random_cvb_init(self):
        K = self._len['K']
        N = self._len['N']
        nfeat = self._len['nfeat']
        dims = self._len['dims']
        zeros = self._len['zeros']
        ones = self._len['ones']
        nnz = self._len['nnz']

        if self._is_symmetric and False:
            self.gamma = np.zeros((N,N,K,K))
            for i, j in self.data_iter():
                gmma = np.random.randint(1, 2*N, (K,K))
                self.frontend.symmetrize(gmma)
                gmma = gmma / gmma.sum()
                self.gamma[i,j] = gmma

                #self.gamma[j,i] = gmma # data_iter 2
                gmma = np.random.randint(1, 2*N, (K,K))
                self.frontend.symmetrize(gmma)
                gmma = gmma / gmma.sum()
                self.gamma[j,i] = gmma
        else:
            self.gamma = np.random.dirichlet(np.ones(K**2)*0.5, (N,N))
            self.gamma.resize(N,N,K,K)
            self.gamma[self.frontend.data_ma == np.ma.masked] = 0 # ???

        #self._symmetric_pt = self._is_symmetric +1
        #self._symmetric_pt = 1

        self.N_theta_left = self.gamma.sum(0).sum(1)
        self.N_theta_right = self.gamma.sum(1).sum(2)

        self.N_phi = np.zeros((nfeat, K, K))
        self.N_phi[0] = self.gamma[self.frontend.data_ma == 0].sum(0)
        self.N_phi[1] = self.gamma[self.frontend.data_ma == 1].sum(0)



    def _reduce_latent(self):
        theta = self.N_theta_right + self.N_theta_left + np.tile(self.hyper_theta, (self.N_theta_left.shape[0],1))
        self._theta = (theta.T / theta.sum(axis=1)).T

        phi = self.N_phi + np.tile(self.hyper_phi, (self.N_phi.shape[1], self.N_phi.shape[2], 1)).T
        #phi = (phi / np.linalg.norm(phi, axis=0))[1]
        self._phi = (phi / phi.sum(0))[1]

        self._K = self.N_phi.shape[1]

        return self._theta, self._phi

    def _reduce_one(self, i, j):
        xij = self._xij

        self.pik = self.N_theta_left[i] + self.hyper_theta
        self.pjk = self.N_theta_right[j] + self.hyper_theta
        pxk = self.N_phi[xij] + self.hyper_phi[xij]

        ##
        outer_kk = np.log(np.outer(self.pik, self.pjk)) + np.log(pxk) - np.log(self.N_phi.sum(0) + self.hyper_phi_sum)
        ##

        return lognormalize(outer_kk.ravel())

    def sample(self):

        expectation_step = self._len['N']/2

        _it = 0
        for i, j in self.data_iter(randomize=True):

            self._xij = self.frontend.data_ma[i,j]
            self.pull_current(i, j)

            qij = self._reduce_one(i, j).reshape(self._len['K'], self._len['K'])

            self.push_current(i, j, qij)

      #      if _it % expectation_step == 0:
      #          self.N_phi[0] = self.gamma[self.frontend.data_ma == 0].sum(0)
      #          self.N_phi[1] = self.gamma[self.frontend.data_ma == 1].sum(0)

      #      _it += 1

      #  self.N_phi[0] = self.gamma[self.frontend.data_ma == 0].sum(0)
      #  self.N_phi[1] = self.gamma[self.frontend.data_ma == 1].sum(0)


    def pull_current(self, i, j):
        xij = self._xij

        q_left = self.gamma[i,j].sum(0)
        q_right = self.gamma[i,j].sum(1)
        self.N_theta_left[i] -= q_left
        self.N_theta_right[j] -= q_right
        self.N_phi[xij] -= self.gamma[i,j]

        if self._is_symmetric:
            self.N_theta_left[j] -= self.gamma[j,i].sum(0)
            self.N_theta_right[i] -= self.gamma[j,i].sum(1)
            self.N_phi[xij] -= self.gamma[j,i]


    def push_current(self, i, j, qij):
        xij = self._xij

        q_left = qij.sum(0)
        q_right = qij.sum(1)
        self.N_theta_left[i] += q_left
        self.N_theta_right[j] += q_right
        self.N_phi[xij] += qij

        self.gamma[i, j] = qij

        if self._is_symmetric:
            self.N_theta_left[j] += q_left
            self.N_theta_right[i] += q_right
            self.N_phi[xij] += qij

            self.gamma[j, i] = qij

    def compute_entropy(self):
        pij = self.likelihood(*self._reduce_latent())

        # Log-likelihood
        pij = self.frontend.data_A * pij + self.frontend.data_B
        ll = np.log(pij).sum()

        # Entropy
        entropy = ll
        #entropy = - ll / self._len['nnz']

        # Perplexity is 2**H(X).

        return entropy

    def compute_entropy_t(self):
        pij = self.likelihood(*self._reduce_latent())

        # Log-likelihood
        pij = self.frontend.data_A_t * pij + self.frontend.data_B_t
        ll = np.log(pij).sum()

        # Entropy
        entropy = ll
        #entropy_t = - ll / self._len['nnz_t']

        # Perplexity is 2**H(X).

        return entropy

    def update_hyper(self, hyper):
        pass


    def generate(self, N=None, K=None, hyperparams=None, mode='predictive', symmetric=True, **kwargs):
        #self.update_hyper(hyperparams)
        #alpha, gmma, delta = self.get_hyper()

        # predictive
        try: theta, phi = self.get_params()
        except: return self.generate(N, K, hyperparams, 'generative', symmetric)
        K = theta.shape[1]

        pij = self.likelihood(theta, phi)
        pij = np.clip(pij, 0, 1)
        Y = sp.stats.bernoulli.rvs(pij)

        return Y


if __name__ == "__main__":

    import pymake
    from pymake import frontendNetwork

    data = np.arange(16).reshape(4,4)
    data_ma = np.ma.array(data, mask = data*0)

    data_ma.mask[0,1] = True
    data_ma.mask[1,1] = True
    data_ma.mask[1,2] = True
    data_ma.mask[1,3] = True
    data_ma.mask[3,3] = True

    fr = frontendNetwork.from_array(data_ma)

    model = immsb_scvb({}, fr)

    _abc = model.data_iter(data_ma)

    # data to iter
    print(data_ma[zip(*_abc)])




from pymake import ExpeFormat

import os
from subprocess import call




USAGE = """\
----------------
Manage the data : This script is part of the repo/ml of pymake.
----------------
"""



class Process(ExpeFormat):

    _default_expe = { '_expe_silent' : True }


    def killall(self):

        cmd = 'killall -9 -TERM pmk'
        call(cmd.split())



# -*- coding: utf-8 -*-


from ml.model.hdp.lda import GibbsRun, Likelihood, ZSampler,  NP_CGS
from ml.model.hdp.hdp import MSampler, BetaSampler

# @idem than immsb_cgs
class ilda_cgs(GibbsRun):
    def __init__(self, expe, frontend):

        delta = expe.hyperparams.get('delta',1)
        alpha = expe.hyperparams.get('alpha',1)
        gmma = expe.hyperparams.get('gmma',1)

        hyper = expe.hyper
        assortativity = expe.get('homo')
        hyper_prior = expe.get('hyper_prior') # HDP hyper optimization
        K = expe.K

        data = frontend.data
        data_t = frontend.data_t

        likelihood = Likelihood(delta, data, assortativity=assortativity)

        # Nonparametric case
        zsampler = ZSampler(alpha, likelihood, K_init=K, data_t=data_t)
        msampler = MSampler(zsampler)
        betasampler = BetaSampler(gmma, msampler)
        jointsampler = NP_CGS(zsampler, msampler, betasampler,
                              hyper=hyper, hyper_prior=hyper_prior)

        super(ilda_cgs, self).__init__(expe, frontend,
                                       jointsampler,
                                       iterations=expe.iterations,
                                       output_path=expe._output_path,
                                       write=expe._write,
                                       data_t=data_t)
        self.update_hyper(expe.hyperparams)





"""
@author: Ke Zhai (zhaike@cs.umd.edu)

This code was modified from the code originally written by Frank Wood (fwood@stat.columbia.edu).
Implements collapsed Gibbs sampling for the infinite Gaussian mixture model (IGMM).
"""

import numpy, scipy;

# We will be taking log(0) = -Inf, so turn off this warning
numpy.seterr(divide='ignore')

class CollapsedGibbsSampling(object):
    import scipy.stats;

    """
    @param truncation_level: the maximum number of clusters, used for speeding up the computation
    @param snapshot_interval: the interval for exporting a snapshot of the model
    """
    def __init__(self,
                 truncation_level=100,
                 snapshot_interval=100):
        self._truncation_level = truncation_level;
        self._snapshot_interval = snapshot_interval;

        self._label_title = "Label-";
        self._mu_title = "Mu-";
        self._sigma_title = "Sigma-";
        self._hyper_parameter_vector_title = "Hyper-parameter-vector-";
        self._hyper_parameter_matrix_title = "Hyper-parameter-matrix-";

    """
    @param data: a N-by-D numpy array object, defines N points of D dimension
    @param alpha: the concentration parameter of the dirichlet process
    @param kappa_0: initial kappa_0
    @param nu_0: initial nu_0
    @param mu_0: initial cluster center
    @param lambda_0: initial lambda_0
    """
    def _initialize(self, data, alpha=1., kappa_0=1., nu_0=1., mu_0=None, lambda_0=None):
        self._X = data;
        (self._N, self._D) = self._X.shape;

        # initialize every point to one cluster
        self._K = 1;
        self._count = numpy.zeros(self._truncation_level, numpy.uint8);
        self._count[0] = self._N;
        self._label = numpy.zeros(self._N, numpy.uint8);

        # compute the sum and square sum of all cluster up to truncation level
        self._sum = numpy.zeros((self._truncation_level, self._D));
        self._sum[0, :] = numpy.sum(self._X, 0);
        self._square_sum = numpy.zeros((self._truncation_level, self._D, self._D));
        self._square_sum[[0], :, :] = numpy.dot(self._X.transpose(), self._X);

        # initialize the initial mean for the cluster
        if mu_0 == None:
            self._mu_0 = numpy.zeros((1, self._D));
        else:
            self._mu_0 = mu_0;
        assert(self._mu_0.shape == (1, self._D));

        # initialize the concentration parameter of the dirichlet distirbution
        self._alpha = alpha;

        # initialize the mean fraction
        self._kappa_0 = kappa_0;

        # initialize the degree of freedom
        self._nu_0 = nu_0;
        if self._nu_0 < self._D:
            print("warning: nu_0 is less than data dimensionality, will set to dimensionality...")
            self._nu_0 = self._D;

        # initialize the lambda
        if lambda_0 == None:
            self._lambda_0 = numpy.eye(self._D);
        else:
            self._lambda_0 = lambda_0;
        assert(self._lambda_0.shape == (self._D, self._D));

        # initialize the sigma, inv(sigma) and log(det(sigma)) of all cluster up to truncation level
        #self._sigma = numpy.zeros((self._truncation_level, self._D, self._D));
        self._sigma_inv = numpy.zeros((self._truncation_level, self._D, self._D));
        self._log_sigma_det = numpy.zeros(self._truncation_level);

        # compute the sigma, inv(sigma) and log(det(sigma)) of the first cluster
        nu = self._nu_0 + self._count[0] - self._D + 1;
        mu_x = self._sum[[0], :] / self._count[0];
        assert(numpy.dot(mu_x.transpose(), mu_x).shape == (self._D, self._D));
        S = self._square_sum[0, :, :] - self._count[0] * numpy.dot(mu_x.transpose(), mu_x);
        mu_x_mu_0 = mu_x - self._mu_0;
        assert(numpy.dot(mu_x_mu_0.transpose(), mu_x_mu_0).shape == (self._D, self._D));
        lambda_n = self._lambda_0 + S + self._kappa_0 * self._count[0] * numpy.dot(mu_x_mu_0.transpose(), mu_x_mu_0) / (self._kappa_0 + self._count[0]);
        sigma = (lambda_n * (self._kappa_0 + self._count[0] + 1)) / ((self._kappa_0 + self._count[0]) * nu);
        assert(numpy.linalg.det(sigma) > 0);

        #self._sigma[0, :, :] = sigma;
        self._sigma_inv[0, :, :] = numpy.linalg.inv(sigma);
        self._log_sigma_det[0] = numpy.log(numpy.linalg.det(sigma));

        # initialize the default log(det(sigma)) and inv(sigma) for new cluster
        nu = self._nu_0 - self._D + 1;
        sigma_0 = self._lambda_0 * (self._kappa_0 + 1) / (self._kappa_0 * nu);
        self._log_sigma_det_0 = numpy.log(numpy.linalg.det(sigma_0));
        self._sigma_inv_0 = numpy.linalg.inv(sigma_0);

    """
    sample the data to train the parameters
    @param iteration: the number of gibbs sampling iteration
    @param directory: the directory to save output, default to "../../output/tmp-output"
    """
    def sample(self, iteration, directory="../../output/tmp-output/"):
        #sample the total data
        for iter in xrange(iteration):
            order = numpy.random.permutation(self._N);
            for (point_counter, point_index) in enumerate(order):
                # get the old label of current point
                old_label = self._label[point_index];
                assert(old_label < self._K and old_label >= 0 and numpy.dtype(old_label) == numpy.uint8);

                # record down the inv(sigma) and log(det(sigma)) of the old cluster
                old_sigma_inv = self._sigma_inv[old_label, :, :];
                old_log_sigma_det = self._log_sigma_det[old_label];

                # remove the current point from the cluster
                self._count[old_label] -= 1;
                if self._count[old_label] == 0:
                    # if current point is from a singleton cluster, shift the last cluster to current one
                    self._count[old_label] = self._count[self._K - 1];
                    self._label[numpy.nonzero(self._label == (self._K - 1))] = old_label;
                    self._sum[old_label, :] = self._sum[self._K - 1, :];
                    self._square_sum[old_label, :, :] = self._square_sum[self._K - 1, :, :];
                    self._K -= 1;
                    # remove the last one to remain compact cluster
                    self._count[self._K] = 0;
                    self._sum[[self._K], :] = numpy.zeros((1, self._D));
                    self._square_sum[[self._K], :, :] = numpy.zeros((1, self._D, self._D));
                else:
                    # change the sum and square sum of the old cluster
                    self._sum[old_label, :] -= self._X[point_index, :];
                    self._square_sum[old_label, :, :] -= numpy.dot(self._X[[point_index], :].transpose(), self._X[[point_index], :]);

                    # change the inv(sigma) and log(det(sigma)) of the old cluster
                    mu_y = self._sum[[old_label], :] / self._count[old_label];
                    kappa_n = self._kappa_0 + self._count[old_label];
                    nu = self._nu_0 + self._count[old_label] - self._D + 1;
                    assert(numpy.dot(mu_y.transpose(), mu_y).shape == (self._D, self._D));
                    S = self._square_sum[old_label, :, :] - self._count[old_label] * numpy.dot(mu_y.transpose(), mu_y);
                    mu_y_mu_0 = mu_y - self._mu_0;
                    assert(numpy.dot(mu_y_mu_0.transpose(), mu_y_mu_0).shape == (self._D, self._D));
                    lambda_n = self._lambda_0 + S + self._kappa_0 * self._count[old_label] * numpy.dot(mu_y_mu_0.transpose(), mu_y_mu_0) / (self._kappa_0 + self._count[old_label]);
                    sigma = (lambda_n * (kappa_n + 1)) / (kappa_n * nu);
                    assert(numpy.linalg.det(sigma) > 0);

                    #self._sigma[old_label, :, :] = sigma;
                    self._sigma_inv[old_label, :, :] = numpy.linalg.inv(sigma);
                    self._log_sigma_det[old_label] = numpy.linalg.det(sigma);

                # compute the prior of being in any of the clusters
                cluster_prior = numpy.hstack((self._count[:self._K], self._alpha));
                cluster_prior = cluster_prior / (self._N - 1. + self._alpha);

                # initialize the likelihood vector for all clusters
                cluster_likelihood = numpy.zeros(self._K + 1);
                # compute the likelihood for the existing clusters
                for k in xrange(self._K):
                    mu_y = self._sum[[k], :] / self._count[k];
                    nu = self._nu_0 + self._count[k] - self._D + 1;
                    mu_n = (self._kappa_0 * self._mu_0 + self._count[k] * mu_y) / (self._kappa_0 + self._count[k]);
                    y_mu_n = self._X[[point_index], :] - mu_n;
                    log_prob = scipy.special.gammaln((nu + self._D) / 2.);
                    log_prob -= (scipy.special.gammaln(nu / 2.) + (self._D / 2.) * numpy.log(nu) + (self._D / 2.) * numpy.log(numpy.pi));
                    log_prob -= 0.5 * self._log_sigma_det[k];
                    assert(numpy.dot(y_mu_n, self._sigma_inv[k, :, :]).shape == (1, self._D));
                    assert(numpy.dot(numpy.dot(y_mu_n, self._sigma_inv[k, :, :]), y_mu_n.transpose()).shape == (1, 1))
                    log_prob -= 0.5 * (nu + self._D) * numpy.log(1. + 1. / nu * numpy.dot(numpy.dot(y_mu_n, self._sigma_inv[k, :, :]), y_mu_n.transpose()));
                    cluster_likelihood[k] = numpy.exp(log_prob);

                # compute the likelihood for new cluster
                nu = self._nu_0 - self._D + 1;
                y_mu_0 = self._X[[point_index], :] - self._mu_0;
                log_prob = scipy.special.gammaln((nu + self._D) / 2.);
                log_prob -= (scipy.special.gammaln(nu / 2.) + (self._D / 2.) * numpy.log(nu) + (self._D / 2.) * numpy.log(numpy.pi));
                log_prob -= 0.5 * self._log_sigma_det_0;
                assert(numpy.dot(y_mu_0, self._sigma_inv_0).shape == (1, self._D));
                assert(numpy.dot(numpy.dot(y_mu_0, self._sigma_inv_0), y_mu_0.transpose()).shape == (1, 1));
                log_prob -= 0.5 * (nu + self._D) * numpy.log(1. + 1. / nu * numpy.dot(numpy.dot(y_mu_0, self._sigma_inv_0), y_mu_0.transpose()));
                cluster_likelihood[self._K] = numpy.exp(log_prob);

                cluster_posterior = cluster_prior * cluster_likelihood;
                cluster_posterior /= numpy.sum(cluster_posterior);

                # sample a new cluster label for current point
                cdf = numpy.cumsum(cluster_posterior);
                new_label = numpy.uint8(numpy.nonzero(cdf >= numpy.random.random())[0][0]);
                assert(new_label >= 0 and new_label <= self._K and numpy.dtype(new_label) == numpy.uint8);

                # if this point starts up a new cluster
                if new_label == self._K:
                    self._K += 1;
                self._count[new_label] += 1;
                self._sum[new_label, :] += self._X[point_index, :];
                self._square_sum[new_label, :, :] += numpy.dot(self._X[[point_index], :].transpose(), self._X[[point_index], :]);
                self._label[point_index] = new_label;

                if old_label == new_label:
                    # if the point is allocated to the old cluster, retrieve all previous parameter
                    self._sigma_inv[new_label, :, :] = old_sigma_inv;
                    self._log_sigma_det[new_label] = old_log_sigma_det;
                else:
                    # if the point is allocated to a new cluster, compute all new parameter
                    mu_y = self._sum[[new_label], :] / self._count[new_label];
                    kappa_n = self._kappa_0 + self._count[new_label];
                    nu = self._nu_0 + self._count[new_label] - self._D + 1;
                    assert(numpy.dot(mu_y.transpose(), mu_y).shape == (self._D, self._D));
                    S = self._square_sum[new_label, :, :] - self._count[new_label] * numpy.dot(mu_y.transpose(), mu_y);
                    mu_y_mu_0 = mu_y - self._mu_0;
                    assert(numpy.dot(mu_y_mu_0.transpose(), mu_y_mu_0).shape == (self._D, self._D));
                    lambda_n = self._lambda_0 + S + self._kappa_0 * self._count[new_label] * numpy.dot(mu_y_mu_0.transpose(), mu_y_mu_0) / (self._kappa_0 + self._count[new_label]);
                    sigma = (lambda_n * (kappa_n + 1)) / (kappa_n * nu);
                    assert(numpy.linalg.det(sigma) > 0);

                    #self._sigma[new_label, :, :] = sigma;
                    self._sigma_inv[new_label, :, :] = numpy.linalg.inv(sigma);
                    self._log_sigma_det[new_label] = numpy.linalg.det(sigma);

            if iter > 0 and iter % 100 == 0:
                print("sampling in progress %2d%%" % (100 * iter / iteration))
                print("total number of cluster %i, likelihood is %f" % (self._K, self.log_likelihood()))

            if (iter + 1) % self._snapshot_interval == 0:
                self.export_snapshot(directory, iter + 1);

    """
    """
    def log_likelihood(self):
        log_likelihood = 0.;
        for n in xrange(self._N):
            log_likelihood -= 0.5 * self._D * numpy.log(2.0 * numpy.pi) + 0.5 * self._log_sigma_det[self._label[n]];
            mean_offset = self._X[n, :][numpy.newaxis, :] - self._sum[self._label[n], :][numpy.newaxis, :]/self._count[self._label[n]];
            assert(mean_offset.shape==(1, self._D));
            log_likelihood -= 0.5 * numpy.dot(numpy.dot(mean_offset, self._sigma_inv[self._label[n], :, :]), mean_offset.transpose());

        #todo: add in the likelihood for K and hyperparameter

        return log_likelihood;

    """
    """
    def export_snapshot(self, directory, index):
        import os
        if not os.path.exists(directory):
            os.mkdir(directory);
        assert(directory.endswith("/"));

        numpy.savetxt(directory + self._label_title + str(index), numpy.uint8(self._label));
        numpy.savetxt(directory + self._mu_title + str(index), self._sum[:self._K, :] / self._count[:self._K][numpy.newaxis, :].transpose());
        sigma = self._sigma_inv;
        for k in xrange(self._K):
            sigma[k, :, :] = numpy.linalg.inv(sigma[k, :, :]);
        numpy.savetxt(directory + self._sigma_title + str(index), numpy.reshape(sigma[:self._K, :, :], (self._K, self._D * self._D)));
        vector = numpy.array([self._alpha, self._kappa_0, self._nu_0]);
        numpy.savetxt(directory + self._hyper_parameter_vector_title + str(index), vector);
        matrix = numpy.vstack((self._mu_0, self._lambda_0));
        numpy.savetxt(directory + self._hyper_parameter_matrix_title, matrix);

        prin( "successfully export the snapshot to " + directory + " for iteration " + str(index) + "...")

"""
run IGMM on the synthetic clustering dataset.
"""
if __name__ == '__main__':
    data = numpy.loadtxt('../../data/clustering/test.dat');

    gs = CollapsedGibbsSampling();
    gs._initialize(data);

    gs.sample(1000);

    print( gs._K)
    print( gs._count[:gs._K])
    print( gs._label)
